# -*- coding: utf-8 -*-
import os

import pandas as pd

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
Radar_info_Path = os.path.join(ROOT_DIR, "data", "radar_info.json")
CN_shp_path = os.path.join(ROOT_DIR, "data", "CHN_pronvices.shp")
last_open_dir = os.path.join(ROOT_DIR, "data", "default_opendir.json")

radar_info = pd.read_json(Radar_info_Path)
_CN_SHP_INFO = None


def get_cn_shp_info():
    """Return the packaged province shapefile reader when cartopy is installed."""
    global _CN_SHP_INFO
    if _CN_SHP_INFO is None:
        try:
            import cartopy.io.shapereader as shpreader
        except ImportError as exc:
            raise ImportError(
                "Map plotting requires cartopy. Install the full optional stack with "
                "`pip install \"pycwr[full]\"`."
            ) from exc
        _CN_SHP_INFO = shpreader.Reader(CN_shp_path)
    return _CN_SHP_INFO


def __getattr__(name):
    if name == "CN_shp_info":
        return get_cn_shp_info()
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# -*- coding: utf-8 -*-
"""Write PRD volumes to NEXRAD Level II archive formats."""

from __future__ import annotations

import datetime
import os
import re
import struct
from collections import OrderedDict

import numpy as np


RECORD_SIZE = 2432
MSG_HEADER_SIZE = 16
MSG31_HEADER_SIZE = 72
VOLUME_HEADER_SIZE = 24
COMPRESSION_RECORD_SIZE = 12

VOLUME_HEADER_FMT = ">9s3sII4s"
MSG1_HEADER_FMT = (
    ">I H h H H H H H "
    "H H H H H H H f "
    "H H H H H "
    "8s 2s 2s 2s "
    "h h h H "
    "32s"
)
MSG1_HEADER_SIZE = struct.calcsize(MSG1_HEADER_FMT)


NEXRAD_MSG31_FIELD_SPECS = OrderedDict(
    [
        ("dBZ", {"moment": b"REF", "scale": 2.0, "offset": 66.0, "word_size": 8}),
        ("V", {"moment": b"VEL", "scale": 2.0, "offset": 129.0, "word_size": 8}),
        ("W", {"moment": b"SW ", "scale": 2.0, "offset": 129.0, "word_size": 8}),
        ("ZDR", {"moment": b"ZDR", "scale": 16.0, "offset": 128.0, "word_size": 8}),
        ("PhiDP", {"moment": b"PHI", "scale": 100.0, "offset": 2.0, "word_size": 16}),
        ("CC", {"moment": b"RHO", "scale": 200.0, "offset": 5.0, "word_size": 8}),
    ]
)
NEXRAD_MSG1_FIELD_SPECS = OrderedDict(
    [
        ("dBZ", {"scale": 2.0, "offset": 66.0}),
        ("V", {"scale": 2.0, "offset": 129.0}),
        ("W", {"scale": 2.0, "offset": 129.0}),
    ]
)


def _ensure_ppi_volume(prd):
    scan_type = str(np.asarray(prd.scan_info["scan_type"].values).item())
    if scan_type != "ppi":
        raise ValueError("NEXRAD Level II export only supports ppi PRD volumes.")


def _ensure_output_path(filename, overwrite=False):
    path = os.path.abspath(filename)
    if os.path.exists(path) and not overwrite:
        raise FileExistsError("Output file already exists: %s" % path)
    parent = os.path.dirname(path)
    if parent:
        os.makedirs(parent, exist_ok=True)
    return path


def _python_datetime(value):
    array = np.asarray(value)
    if np.issubdtype(array.dtype, np.datetime64):
        return array.astype("datetime64[ms]").tolist()
    if isinstance(value, datetime.datetime):
        return value
    raise TypeError("Unable to convert %r to datetime." % (value,))


def _compute_mjd(dt):
    return int((dt.date() - datetime.date(1970, 1, 1)).days + 1)


def _compute_milliseconds(dt):
    midnight = datetime.datetime(dt.year, dt.month, dt.day, tzinfo=dt.tzinfo)
    return int((dt - midnight).total_seconds() * 1000)


def _normalize_icao(icao):
    if icao is None:
        return b"PYCW"
    text = str(icao).strip().upper()
    if not text:
        return b"PYCW"
    return text[:4].ljust(4).encode("ascii", "ignore")


def _infer_site_code(prd, fallback="PYCW"):
    sitename = str(getattr(prd, "sitename", "") or "")
    match = re.search(r"([A-Z0-9]{4,8})$", sitename.upper())
    if match is not None:
        return match.group(1)
    return fallback


def _normalize_station_bytes(text, length, fallback):
    raw = (str(text).strip() if text is not None else "") or fallback
    return raw[:length].ljust(length).encode("ascii", "ignore")


def _available_supported_fields(prd, supported_specs, field_names=None, strict=True, required_fields=None):
    available = set(prd.available_fields(range_mode=None))
    def _is_resolvable(field_name):
        if hasattr(prd, "resolve_field_name"):
            return prd.resolve_field_name(field_name, range_mode=None, required=False) is not None
        return field_name in available
    required = tuple(required_fields or ())
    if field_names is None:
        selected = [name for name in supported_specs if _is_resolvable(name)]
    else:
        selected = []
        for name in field_names:
            if name not in supported_specs:
                if strict:
                    raise ValueError("Field %s is not supported by this export format." % name)
                continue
            if not _is_resolvable(name):
                if strict:
                    raise KeyError(name)
                continue
            selected.append(name)
    for name in required:
        if name not in selected:
            if field_names is None:
                raise ValueError("Field %s is required by this export format." % name)
            raise ValueError("Field %s is required by this export format." % name)
    if not selected:
        raise ValueError("No exportable fields are available for this format.")
    return selected


def _sweep_state(isweep, iray, rays_per_sweep, nsweeps):
    if isweep == 0 and iray == 0:
        return 3
    if isweep == nsweeps - 1 and iray == rays_per_sweep - 1:
        return 4
    if iray == 0:
        return 0
    if iray == rays_per_sweep - 1:
        return 2
    return 1


def _get_prd_ray(prd, sweep, field_name, iray):
    source_name = field_name
    if hasattr(prd, "resolve_field_name"):
        source_name = prd.resolve_field_name(field_name, sweep=sweep, range_mode=None, required=False)
    if source_name is None:
        raise KeyError(field_name)
    field = prd.get_sweep_field(sweep, source_name, range_mode=None)
    values = np.asarray(field.values[iray], dtype=np.float32)
    ranges = np.asarray(field["range"].values, dtype=np.float64)
    return values, ranges


def _range_geometry(range_values):
    range_values = np.asarray(range_values, dtype=np.float64)
    if range_values.size == 0:
        raise ValueError("Export range vector cannot be empty.")
    if range_values.size > 1:
        spacing = float(range_values[1] - range_values[0])
        if spacing <= 0.0:
            raise ValueError("Range spacing must be positive.")
    else:
        spacing = float(range_values[0]) if float(range_values[0]) > 0.0 else 1.0
    return int(round(float(range_values[0]))), int(round(spacing))


def _encode_quantized(values, scale, offset, max_code, missing_code=0, word_size=8):
    codes = np.round(np.asarray(values, dtype=np.float64) * float(scale) + float(offset))
    codes[~np.isfinite(values)] = missing_code
    codes = np.clip(codes.astype(np.int64), 0, max_code)
    if word_size == 8:
        return codes.astype(np.uint8).tobytes()
    if word_size == 16:
        return codes.astype(">u2").tobytes()
    raise ValueError("Unsupported word size: %s" % word_size)


def _pack_msg31_moment_block(moment_name, values, first_gate, gate_spacing, scale, offset, word_size):
    encoded = _encode_quantized(values, scale, offset, 255 if word_size == 8 else 65535, word_size=word_size)
    header = struct.pack(
        ">1s3sI H h h h h B B f f",
        b"B",
        moment_name,
        0,
        len(values),
        first_gate,
        gate_spacing,
        0,
        0,
        0,
        word_size,
        float(scale),
        float(offset),
    )
    block = header + encoded
    if len(block) % 2:
        block += b"\x00"
    return block


def _pack_msg31_radial(prd, sweep, iray, field_names, seq_id):
    azimuth = float(prd.fields[sweep].azimuth.values[iray])
    elevation = float(prd.fields[sweep].elevation.values[iray])
    dt = _python_datetime(prd.fields[sweep].time.values[iray])
    mjd = _compute_mjd(dt)
    milliseconds = _compute_milliseconds(dt)
    rays_per_sweep = int(prd.scan_info["rays_per_sweep"].values[sweep])
    radial_state = _sweep_state(sweep, iray, rays_per_sweep, int(prd.nsweeps))
    nyquist = float(prd.scan_info["nyquist_velocity"].values[sweep]) if "nyquist_velocity" in prd.scan_info else 0.0
    altitude = int(round(float(prd.scan_info["altitude"].values)))
    lat = float(prd.scan_info["latitude"].values)
    lon = float(prd.scan_info["longitude"].values)
    vcp = int(np.asarray(prd.metadata.get("volume_number", 0)))

    moment_blocks = []
    max_range_m = 0.0
    for field_name in field_names:
        spec = NEXRAD_MSG31_FIELD_SPECS[field_name]
        values, ranges = _get_prd_ray(prd, sweep, field_name, iray)
        first_gate, gate_spacing = _range_geometry(ranges)
        if ranges.size:
            max_range_m = max(max_range_m, float(ranges[-1]))
        moment_blocks.append(
            _pack_msg31_moment_block(
                spec["moment"],
                values,
                first_gate,
                gate_spacing,
                spec["scale"],
                spec["offset"],
                spec["word_size"],
            )
        )

    msg_header = struct.pack(
        ">HBBHHIHH",
        0,
        1,
        31,
        seq_id & 0x7FFF,
        mjd,
        milliseconds,
        1,
        1,
    )

    ptr_vol = MSG31_HEADER_SIZE
    ptr_elv = ptr_vol + 44
    ptr_rad = ptr_elv + 12
    pointers = [ptr_vol, ptr_elv, ptr_rad]
    next_offset = ptr_rad + 20
    for block in moment_blocks:
        pointers.append(next_offset)
        next_offset += len(block)
    full_pointers = pointers[:10] + [0] * max(0, 10 - len(pointers))

    radial_length = 44 + 12 + 20 + sum(len(block) for block in moment_blocks)
    msg31_header = struct.pack(
        ">4s I H H f B B H B B B B f B b H " + "I" * 10,
        b"RAD1",
        milliseconds,
        mjd,
        iray + 1,
        azimuth,
        0,
        0,
        radial_length,
        0,
        0,
        sweep + 1,
        radial_state,
        elevation,
        0,
        0,
        3 + len(moment_blocks),
        *full_pointers[:10],
    )

    vol_block = struct.pack(
        ">1s3sHBBffhhfffffH2s",
        b"V",
        b"VOL",
        40,
        1,
        0,
        lat,
        lon,
        altitude,
        0,
        0.0,
        0.0,
        0.0,
        0.0,
        0.0,
        max(0, min(65535, vcp)),
        b"\x00\x00",
    )
    elv_block = struct.pack(">1s3sHhf", b"E", b"ELV", 12, 0, elevation)
    radial_block = struct.pack(
        ">1s3sHhffh2s",
        b"R",
        b"RAD",
        20,
        int(round(max_range_m / 100.0)),
        0.0,
        0.0,
        int(round(nyquist * 100.0)),
        b"\x00\x00",
    )

    radial = bytearray()
    radial += msg_header
    radial += msg31_header
    radial += vol_block
    radial += elv_block
    radial += radial_block
    for block in moment_blocks:
        radial += block
    radial += struct.pack(">I", 0xFFFFFFFF)
    radial_size_halfwords = (len(radial) - MSG_HEADER_SIZE + 4) // 2
    radial[0:2] = struct.pack(">H", radial_size_halfwords)
    return bytes(radial), (seq_id + 1) & 0x7FFF


def _pack_msg5(prd, icao):
    now = _python_datetime(prd.scan_info["start_time"].values)
    mjd = _compute_mjd(now)
    milliseconds = _compute_milliseconds(now)
    msg_header = struct.pack(
        ">HBBHHIHH",
        0,
        1,
        5,
        0,
        mjd,
        milliseconds,
        1,
        1,
    )
    msg5_main = struct.pack(
        ">H H H H H B B 10s",
        0,
        1,
        0,
        int(prd.nsweeps),
        0,
        2,
        2,
        b"\x00" * 10,
    )
    elev_blocks = bytearray()
    fixed_angles = np.asarray(prd.scan_info["fixed_angle"].values, dtype=np.float64)
    for angle in fixed_angles:
        elev_blocks += struct.pack(
            ">HBBBBHHhhhhhhHHH2sHHH2sHHH2s",
            int(round(float(angle) * 65536.0 / 360.0)),
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            b"\x00\x00",
            0,
            0,
            0,
            b"\x00\x00",
            0,
            0,
            0,
            b"\x00\x00",
        )
    msg5 = bytearray()
    msg5 += msg_header
    msg5 += msg5_main
    msg5 += elev_blocks
    msg5[0:2] = struct.pack(">H", (len(msg5) - 2) // 2)
    if len(msg5) > RECORD_SIZE:
        raise ValueError("MSG5 record exceeds the fixed NEXRAD record size.")
    msg5 += b"\x00" * (RECORD_SIZE - len(msg5))
    return bytes(msg5)


def write_nexrad_level2_msg31(prd, filename, field_names=None, strict=True, overwrite=False, icao=None):
    _ensure_ppi_volume(prd)
    path = _ensure_output_path(filename, overwrite=overwrite)
    selected_fields = _available_supported_fields(prd, NEXRAD_MSG31_FIELD_SPECS, field_names=field_names, strict=strict)
    start_time = _python_datetime(prd.scan_info["start_time"].values)
    volume_header = struct.pack(
        VOLUME_HEADER_FMT,
        b"AR2V0006.",
        b"001",
        _compute_mjd(start_time),
        _compute_milliseconds(start_time),
        _normalize_icao(icao),
    )
    seq_id = 0
    with open(path, "wb") as handle:
        handle.write(volume_header)
        handle.write(b"\x00" * COMPRESSION_RECORD_SIZE)
        handle.write(_pack_msg5(prd, icao))
        for sweep in range(int(prd.nsweeps)):
            rays_per_sweep = int(prd.scan_info["rays_per_sweep"].values[sweep])
            for iray in range(rays_per_sweep):
                radial, seq_id = _pack_msg31_radial(prd, sweep, iray, selected_fields, seq_id)
                handle.write(radial)
    return path


def _pack_msg1_radial(prd, sweep, iray, seq_id):
    rays_per_sweep = int(prd.scan_info["rays_per_sweep"].values[sweep])
    radial_state = _sweep_state(sweep, iray, rays_per_sweep, int(prd.nsweeps))
    dt = _python_datetime(prd.fields[sweep].time.values[iray])
    mjd = _compute_mjd(dt)
    milliseconds = _compute_milliseconds(dt)
    nyquist = float(prd.scan_info["nyquist_velocity"].values[sweep]) if "nyquist_velocity" in prd.scan_info else 0.0
    unambiguous_range = float(prd.scan_info["unambiguous_range"].values[sweep]) if "unambiguous_range" in prd.scan_info else 0.0
    azimuth = float(prd.fields[sweep].azimuth.values[iray])
    elevation = float(prd.fields[sweep].elevation.values[iray])

    ref_values, ref_ranges = _get_prd_ray(prd, sweep, "dBZ", iray)
    vel_values, vel_ranges = _get_prd_ray(prd, sweep, "V", iray)
    sw_values, sw_ranges = _get_prd_ray(prd, sweep, "W", iray)
    first_gate, gate_spacing = _range_geometry(ref_ranges)
    max_gates = (RECORD_SIZE - MSG_HEADER_SIZE - MSG1_HEADER_SIZE - 4) // 3
    ngates = min(len(ref_values), len(vel_values), len(sw_values), max_gates)
    ref_values = ref_values[:ngates]
    vel_values = vel_values[:ngates]
    sw_values = sw_values[:ngates]

    msg_header = struct.pack(
        ">HBBHHIHH",
        1214,
        1,
        1,
        seq_id & 0x7FFF,
        mjd,
        milliseconds,
        1,
        1,
    )
    angle_scale = 180.0 / (4096.0 * 8.0)
    msg1_header = struct.pack(
        MSG1_HEADER_FMT,
        milliseconds,
        mjd,
        int(round(unambiguous_range / 100.0)),
        max(0, min(0xFFFF, int(round(azimuth / angle_scale)))),
        iray + 1,
        radial_state,
        max(0, min(0xFFFF, int(round(elevation / angle_scale)))),
        sweep + 1,
        first_gate,
        first_gate,
        gate_spacing,
        gate_spacing,
        ngates,
        ngates,
        0,
        0.0,
        MSG1_HEADER_SIZE,
        MSG1_HEADER_SIZE + ngates,
        MSG1_HEADER_SIZE + 2 * ngates,
        2,
        0,
        b"\x00" * 8,
        b"\x00" * 2,
        b"\x00" * 2,
        b"\x00" * 2,
        int(round(nyquist * 100.0)),
        0,
        0,
        0,
        b"\x00" * 32,
    )
    radial = bytearray()
    radial += msg_header
    radial += msg1_header
    radial += _encode_quantized(ref_values, 2.0, 66.0, 255, word_size=8)
    radial += _encode_quantized(vel_values, 2.0, 129.0, 255, word_size=8)
    radial += _encode_quantized(sw_values, 2.0, 129.0, 255, word_size=8)
    radial += struct.pack(">I", 0xFFFFFFFF)
    if len(radial) > RECORD_SIZE:
        raise ValueError("MSG1 radial exceeds fixed record size.")
    radial += b"\x00" * (RECORD_SIZE - len(radial))
    return bytes(radial), (seq_id + 1) & 0x7FFF


def write_nexrad_level2_msg1(prd, filename, field_names=None, strict=True, overwrite=False, icao=None):
    _ensure_ppi_volume(prd)
    _available_supported_fields(
        prd,
        NEXRAD_MSG1_FIELD_SPECS,
        field_names=field_names,
        strict=strict,
        required_fields=("dBZ", "V", "W"),
    )
    path = _ensure_output_path(filename, overwrite=overwrite)
    start_time = _python_datetime(prd.scan_info["start_time"].values)
    volume_header = struct.pack(
        VOLUME_HEADER_FMT,
        b"ARCHIVE2 ",
        b"   ",
        _compute_mjd(start_time),
        _compute_milliseconds(start_time),
        _normalize_icao(icao),
    )
    compression_record = struct.pack(">I", 0) + struct.pack(">H", RECORD_SIZE) + (b"\x00" * 6)
    seq_id = 0
    with open(path, "wb") as handle:
        handle.write(volume_header)
        handle.write(compression_record)
        for sweep in range(int(prd.nsweeps)):
            rays_per_sweep = int(prd.scan_info["rays_per_sweep"].values[sweep])
            for iray in range(rays_per_sweep):
                radial, seq_id = _pack_msg1_radial(prd, sweep, iray, seq_id)
                handle.write(radial)
    return path



from . import NRadar, PyartRadar, transforms, RadarGrid, interop, RadarProduct

try:
    from . import RadarGridC
except ImportError:
    # Fall back to the pure-Python implementation when the optional Cython
    # extension is not available for the current platform or Python ABI.
    RadarGridC = RadarGrid

__all__ = ["NRadar", "PyartRadar", "transforms", "RadarGrid", "RadarGridC", "RadarProduct", "interop"]



# -*- coding: utf-8 -*-
"""
Dual-polarization QC pipeline.
"""

import copy

import numpy as np
import xarray as xr

from ..configure.default_config import CINRAD_field_mapping, DEFAULT_METADATA
from .attenuation import correct_attenuation_kdp, resolve_kdp_coefficients
from .polarimetric import (
    QC_REFERENCE_NOTES,
    build_clear_air_mask,
    build_meteo_mask,
    despeckle_mask,
    kdp_from_phidp,
    phidp_texture,
    smooth_phidp,
)


def _empty_like(reference):
    return np.full_like(np.asanyarray(reference, dtype=float), np.nan, dtype=float)


def _field_metadata(field_name):
    if field_name in CINRAD_field_mapping:
        metadata_key = CINRAD_field_mapping[field_name]
        if metadata_key in DEFAULT_METADATA:
            return dict(DEFAULT_METADATA[metadata_key])
    return {}


def _build_data_array(template, values, field_name):
    data_array = xr.DataArray(
        np.asanyarray(values),
        dims=template.dims,
        coords=template.coords,
        name=field_name,
    )
    data_array.attrs = _field_metadata(field_name)
    return data_array


def run_dualpol_qc(
    ref,
    zdr=None,
    phidp=None,
    kdp=None,
    rhohv=None,
    snr=None,
    dr=0.075,
    band="C",
    use_existing_kdp=True,
    clear_air_mode="label",
    clear_air_max_ref=15.0,
    clear_air_max_rhohv=0.97,
    clear_air_max_phidp_texture=10.0,
    clear_air_max_snr=20.0,
):
    """
    Run a classical dual-polarization QC chain on one sweep array set.
    """
    reflectivity = np.asanyarray(ref, dtype=float)
    if reflectivity.ndim < 2:
        raise ValueError("ref must have at least two dimensions with range on the last axis")

    gate_length = float(dr)
    gamma, beta = resolve_kdp_coefficients(band=band)
    base_mask = np.isfinite(reflectivity)

    if phidp is not None:
        phase = np.asanyarray(phidp, dtype=float)
        phidp_smooth = smooth_phidp(phase, mask=base_mask)
        texture = phidp_texture(phidp_smooth)
    else:
        phase = None
        phidp_smooth = _empty_like(reflectivity)
        texture = _empty_like(reflectivity)

    if clear_air_mode not in {"ignore", "label", "mask"}:
        raise ValueError("clear_air_mode must be one of 'ignore', 'label', or 'mask'")

    meteo_mask = build_meteo_mask(
        reflectivity,
        rhohv=rhohv,
        phidp_texture=None if phase is None else texture,
        snr=snr,
    )
    if clear_air_mode == "ignore":
        clear_air_mask = np.zeros_like(reflectivity, dtype=bool)
    else:
        clear_air_mask = build_clear_air_mask(
            reflectivity,
            rhohv=rhohv,
            phidp_texture=None if phase is None else texture,
            snr=snr,
            max_ref=clear_air_max_ref,
            max_rhohv=clear_air_max_rhohv,
            max_phidp_texture=clear_air_max_phidp_texture,
            max_snr=clear_air_max_snr,
        )
    accepted_mask = meteo_mask & base_mask
    if clear_air_mode == "mask":
        accepted_mask &= ~clear_air_mask
    qc_mask = despeckle_mask(accepted_mask, min_size=9, connectivity=8)

    if kdp is not None and use_existing_kdp:
        kdp_used = np.where(qc_mask, np.asanyarray(kdp, dtype=float), np.nan)
    elif phase is not None:
        kdp_used = np.where(qc_mask, kdp_from_phidp(phidp_smooth, dr=gate_length), np.nan)
    else:
        raise ValueError("dual-pol QC requires either kdp or phidp")

    z_corr, pia, zdr_corr, pia_zdr = correct_attenuation_kdp(
        reflectivity,
        kdp_used,
        dr=gate_length,
        zdr=zdr,
        gamma=gamma,
        beta=beta,
        mask=qc_mask,
    )

    z_corr = np.where(qc_mask, z_corr, np.nan)
    pia = np.where(qc_mask, pia, np.nan)
    kdp_used = np.where(qc_mask, kdp_used, np.nan)

    if zdr is None:
        zdr_corr = _empty_like(reflectivity)
        pia_zdr = _empty_like(reflectivity)
    else:
        zdr_corr = np.where(qc_mask, zdr_corr, np.nan)
        pia_zdr = np.where(qc_mask, pia_zdr, np.nan)

    return {
        "ref_corrected": z_corr,
        "zdr_corrected": zdr_corr,
        "pia": pia,
        "pia_zdr": pia_zdr,
        "phidp_smooth": phidp_smooth,
        "kdp_used": kdp_used,
        "phidp_texture": texture,
        "meteo_mask": meteo_mask,
        "clear_air_mask": clear_air_mask,
        "qc_mask": qc_mask,
        "qc_reference_notes": dict(QC_REFERENCE_NOTES),
    }


def apply_dualpol_qc(
    prd,
    sweeps=None,
    inplace=False,
    band="C",
    use_existing_kdp=True,
    clear_air_mode="label",
    clear_air_max_ref=15.0,
    clear_air_max_rhohv=0.97,
    clear_air_max_phidp_texture=10.0,
    clear_air_max_snr=20.0,
):
    """
    Apply the dual-polarization QC chain to selected sweeps of a PRD object.
    """
    target = prd if inplace else copy.deepcopy(prd)
    if sweeps is None:
        sweeps = list(range(target.nsweeps))

    for sweep in sweeps:
        sweep = int(sweep)
        dataset = target.fields[sweep]
        if "dBZ" not in dataset:
            raise KeyError("Sweep %d is missing dBZ" % sweep)
        if "KDP" not in dataset and "PhiDP" not in dataset:
            raise ValueError("Sweep %d must contain KDP or PhiDP for dual-pol QC" % sweep)

        reference_field = dataset["dBZ"]
        range_values = np.asanyarray(dataset["range"].values, dtype=float)
        if range_values.size < 2:
            raise ValueError("Sweep %d must contain at least two range gates" % sweep)
        gate_length = float(np.median(np.diff(range_values)) / 1000.0)

        results = run_dualpol_qc(
            reference_field.values,
            zdr=None if "ZDR" not in dataset else dataset["ZDR"].values,
            phidp=None if "PhiDP" not in dataset else dataset["PhiDP"].values,
            kdp=None if "KDP" not in dataset else dataset["KDP"].values,
            rhohv=None if "CC" not in dataset else dataset["CC"].values,
            snr=None if "SNRH" not in dataset else dataset["SNRH"].values,
            dr=gate_length,
            band=band,
            use_existing_kdp=use_existing_kdp,
            clear_air_mode=clear_air_mode,
            clear_air_max_ref=clear_air_max_ref,
            clear_air_max_rhohv=clear_air_max_rhohv,
            clear_air_max_phidp_texture=clear_air_max_phidp_texture,
            clear_air_max_snr=clear_air_max_snr,
        )

        dataset["Zc"] = _build_data_array(reference_field, results["ref_corrected"], "Zc")
        dataset["PIA"] = _build_data_array(reference_field, results["pia"], "PIA")
        dataset["PhiDP_smooth"] = _build_data_array(reference_field, results["phidp_smooth"], "PhiDP_smooth")
        dataset["PhiDP_texture"] = _build_data_array(reference_field, results["phidp_texture"], "PhiDP_texture")
        dataset["KDPc"] = _build_data_array(reference_field, results["kdp_used"], "KDPc")
        dataset["METEO_MASK"] = _build_data_array(reference_field, results["meteo_mask"], "METEO_MASK")
        dataset["CLEAR_AIR_MASK"] = _build_data_array(reference_field, results["clear_air_mask"], "CLEAR_AIR_MASK")
        dataset["QC_MASK"] = _build_data_array(reference_field, results["qc_mask"], "QC_MASK")

        if "ZDR" in dataset:
            zdr_template = dataset["ZDR"]
        else:
            zdr_template = reference_field
        dataset["ZDRc"] = _build_data_array(zdr_template, results["zdr_corrected"], "ZDRc")
        dataset["PIA_ZDR"] = _build_data_array(zdr_template, results["pia_zdr"], "PIA_ZDR")

    if hasattr(target, "_invalidate_cached_views"):
        target._invalidate_cached_views()
    return target



from easydict import EasyDict as edict

__C = edict()
cfg = __C

###set interp configure
__C.interp = edict()
__C.interp.mroi = 500 ##最小的影响半径
__C.interp.coeff = 50 ##随着距离圈增长，影响半径变大，增长倍数

###set 3D radar network configure
__C.network = edict()
__C.network.radar_dirs = []
__C.network.field_names = ["dBZ"]
__C.network.field_range_mode = {"dBZ": "native"}
__C.network.lon_min = None
__C.network.lon_max = None
__C.network.lat_min = None
__C.network.lat_max = None
__C.network.lon_res_deg = 0.01
__C.network.lat_res_deg = 0.01
__C.network.level_heights = []
__C.network.product_level_heights = []
__C.network.max_range_km = 460.0
__C.network.time_tolerance_minutes = 10
__C.network.composite_method = "quality_weighted"
__C.network.influence_radius_m = 300000.0
__C.network.quality_weight_terms = ["range", "beam", "vertical", "qc", "blockage"]
__C.network.range_decay_m = 230000.0
__C.network.beam_radius_ref_m = 3000.0
__C.network.vertical_gap_ref_deg = 0.5
__C.network.qc_weight_field = "QC_MASK"
__C.network.blockage_config = None
__C.network.fillvalue = -999.0
__C.network.output_dir = None
__C.network.effective_earth_radius = None
__C.network.file_pattern = "*.bin*"
__C.network.parallel = True
__C.network.max_workers = None
__C.network.blind_method = "hybrid"
__C.network.output_products = ["CR", "VIL", "ET"]
__C.network.plot_overview = True
__C.network.plot_style = "reference"
__C.network.plot_canvas_px = [920, 790]
__C.network.plot_height_levels = []
__C.network.plot_output_dir = None
__C.network.plot_format = "png"
__C.network.plot_product_for_levels = "CAPPI"
__C.network.coverage_policy = "observable_only"
__C.network.vil_min_dbz = 18.0
__C.network.vil_max_dbz_cap = 56.0
__C.network.et_threshold_dbz = 18.0
__C.network.use_qc = False
__C.network.qc_method = "dualpol"
__C.network.qc_band = "C"
__C.network.qc_use_existing_kdp = True
__C.network.qc_fallback = "original"
__C.network.qc_clear_air_mode = "label"
__C.network.qc_clear_air_max_ref = 15.0
__C.network.qc_clear_air_max_rhohv = 0.97
__C.network.qc_clear_air_max_phidp_texture = 10.0
__C.network.qc_clear_air_max_snr = 20.0



"""QC examples and primitive regression tests."""

import unittest

import numpy as np


class AttenuationTests(unittest.TestCase):
    def test_pia_from_kdp_is_monotonic_and_clips_negative_kdp(self):
        from pycwr.qc import pia_from_kdp

        kdp = np.array([[0.5, 1.0, -0.2, 2.0, np.nan]])
        pia = pia_from_kdp(kdp, dr=0.1, gamma=0.08)

        self.assertTrue(np.all(np.diff(pia[0, :4]) >= -1e-12))
        self.assertTrue(np.isnan(pia[0, 4]))

    def test_correct_attenuation_kdp_increases_reflectivity(self):
        from pycwr.qc import correct_attenuation_kdp

        ref = np.array([[20.0, 22.0, 24.0, 26.0]])
        zdr = np.array([[0.5, 0.6, 0.7, 0.8]])
        kdp = np.array([[0.3, 0.4, 0.5, 0.6]])

        zc, pia, zdrc, pia_zdr = correct_attenuation_kdp(ref, kdp, dr=0.1, zdr=zdr)

        self.assertEqual(zc.shape, ref.shape)
        self.assertEqual(pia.shape, ref.shape)
        self.assertTrue(np.all(zc >= ref))
        self.assertTrue(np.all(zdrc >= zdr))
        self.assertTrue(np.all(np.diff(pia[0]) >= -1e-12))
        self.assertTrue(np.all(np.diff(pia_zdr[0]) >= -1e-12))

    def test_correct_attenuation_hb_supports_leading_dimensions(self):
        from pycwr.qc import correct_attenuation_HB

        ref = np.array(
            [
                [[20.0, 21.0, 22.0], [30.0, 31.0, 32.0]],
                [[15.0, 16.0, np.nan], [25.0, 26.0, 27.0]],
            ]
        )
        zc, pia = correct_attenuation_HB(ref, gate_length=0.075)

        self.assertEqual(zc.shape, ref.shape)
        self.assertEqual(pia.shape, ref.shape)
        self.assertTrue(np.isnan(zc[1, 0, 2]))
        self.assertTrue(np.isnan(pia[1, 0, 2]))


class PolarimetricPrimitiveTests(unittest.TestCase):
    def test_smooth_phidp_removes_spike_and_enforces_monotonicity(self):
        from pycwr.qc import smooth_phidp

        phidp = np.array([[0.0, 1.0, 40.0, 3.0, 4.0, 5.0, 6.0]])
        smoothed = smooth_phidp(phidp, median_window=3, fit_window=3, enforce_monotonic=True)

        self.assertLess(smoothed[0, 2], 10.0)
        self.assertTrue(np.all(np.diff(smoothed[0][np.isfinite(smoothed[0])]) >= -1e-12))

    def test_kdp_from_phidp_recovers_linear_slope(self):
        from pycwr.qc import kdp_from_phidp

        dr = 0.1
        kdp_true = 1.0
        gates = np.arange(11)
        phidp = (2.0 * kdp_true * dr) * gates
        kdp = kdp_from_phidp(phidp[np.newaxis, :], dr=dr, fit_window=5)

        self.assertTrue(np.allclose(kdp[0, 2:-2], kdp_true, atol=1e-6))

    def test_phidp_texture_highlights_spikes(self):
        from pycwr.qc import phidp_texture

        phidp = np.array([[1.0, 1.0, 1.0, 10.0, 1.0, 1.0, 1.0]])
        texture = phidp_texture(phidp, window=3)

        self.assertGreater(texture[0, 3], texture[0, 0])

    def test_build_meteo_mask_combines_classic_thresholds(self):
        from pycwr.qc import build_meteo_mask

        ref = np.array([[10.0, 10.0, -5.0]])
        rhohv = np.array([[0.95, 0.60, 0.95]])
        texture = np.array([[5.0, 5.0, 25.0]])
        snr = np.array([[10.0, 10.0, 10.0]])

        mask = build_meteo_mask(ref, rhohv=rhohv, phidp_texture=texture, snr=snr)

        np.testing.assert_array_equal(mask, np.array([[True, False, False]]))

    def test_build_clear_air_mask_identifies_weak_nonprecip_echo(self):
        from pycwr.qc import build_clear_air_mask

        ref = np.array([[8.0, 8.0, 25.0]])
        rhohv = np.array([[0.92, 0.99, 0.99]])
        texture = np.array([[4.0, 4.0, 4.0]])
        snr = np.array([[10.0, 24.0, 24.0]])

        mask = build_clear_air_mask(ref, rhohv=rhohv, phidp_texture=texture, snr=snr)

        np.testing.assert_array_equal(mask, np.array([[True, False, False]]))

    def test_despeckle_mask_removes_small_isolated_echoes(self):
        from pycwr.qc import despeckle_mask

        mask = np.zeros((6, 6), dtype=bool)
        mask[2:5, 2:5] = True
        mask[0, 0] = True

        despeckled = despeckle_mask(mask, min_size=4)

        self.assertFalse(despeckled[0, 0])
        self.assertTrue(np.all(despeckled[2:5, 2:5]))


class PipelineTests(unittest.TestCase):
    def _build_synthetic_prd(self):
        from pycwr.core.NRadar import PRD

        nrays = 4
        nbins = 12
        fields = {
            "dBZ": np.tile(np.linspace(15.0, 35.0, nbins), (nrays, 1)),
            "ZDR": np.tile(np.linspace(0.2, 1.0, nbins), (nrays, 1)),
            "PhiDP": np.tile(np.linspace(0.0, 20.0, nbins), (nrays, 1)),
            "KDP": np.tile(np.linspace(0.3, 1.2, nbins), (nrays, 1)),
            "CC": np.ones((nrays, nbins)) * 0.98,
            "SNRH": np.ones((nrays, nbins)) * 20.0,
        }
        return PRD(
            fields=fields,
            scan_type="ppi",
            time=np.array(
                [np.datetime64("2026-03-20T00:00:00") + np.timedelta64(i, "s") for i in range(nrays)]
            ),
            range=np.arange(nbins, dtype=float) * 75.0,
            azimuth=np.linspace(0.0, 270.0, nrays),
            elevation=np.ones(nrays) * 1.0,
            latitude=30.0,
            longitude=120.0,
            altitude=100.0,
            sweep_start_ray_index=np.array([0]),
            sweep_end_ray_index=np.array([nrays - 1]),
            fixed_angle=np.array([1.0]),
            bins_per_sweep=np.array([nbins]),
            nyquist_velocity=np.array([15.0]),
            frequency=5.6,
            unambiguous_range=np.array([150000.0]),
            nrays=nrays,
            nsweeps=1,
            sitename="TEST",
        )

    def test_run_dualpol_qc_returns_expected_products(self):
        from pycwr.qc import run_dualpol_qc

        ref = np.tile(np.linspace(15.0, 35.0, 12), (4, 1))
        zdr = np.tile(np.linspace(0.2, 1.0, 12), (4, 1))
        phidp = np.tile(np.linspace(0.0, 20.0, 12), (4, 1))
        kdp = np.tile(np.linspace(0.3, 1.2, 12), (4, 1))
        rhohv = np.ones_like(ref) * 0.98
        snr = np.ones_like(ref) * 20.0

        result = run_dualpol_qc(ref, zdr=zdr, phidp=phidp, kdp=kdp, rhohv=rhohv, snr=snr, dr=0.075)

        self.assertEqual(set(result.keys()), {
            "clear_air_mask",
            "ref_corrected",
            "zdr_corrected",
            "pia",
            "pia_zdr",
            "phidp_smooth",
            "kdp_used",
            "phidp_texture",
            "meteo_mask",
            "qc_mask",
            "qc_reference_notes",
        })
        self.assertTrue(np.all(np.isnan(result["ref_corrected"]) | (result["ref_corrected"] >= ref)))

    def test_run_dualpol_qc_can_mask_clear_air_echoes(self):
        from pycwr.qc import run_dualpol_qc

        ref = np.tile(np.array([8.0, 24.0, 26.0, 28.0, 30.0, 32.0, 34.0, 35.0, 36.0, 37.0]), (3, 1))
        zdr = np.tile(np.linspace(0.2, 1.0, 10), (3, 1))
        phidp = np.tile(np.linspace(0.0, 18.0, 10), (3, 1))
        rhohv = np.tile(np.array([0.92, 0.98, 0.98, 0.98, 0.98, 0.98, 0.98, 0.98, 0.98, 0.98]), (3, 1))
        snr = np.tile(np.array([10.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0]), (3, 1))

        labeled = run_dualpol_qc(
            ref,
            zdr=zdr,
            phidp=phidp,
            rhohv=rhohv,
            snr=snr,
            dr=0.075,
            use_existing_kdp=False,
            clear_air_mode="label",
        )
        masked = run_dualpol_qc(
            ref,
            zdr=zdr,
            phidp=phidp,
            rhohv=rhohv,
            snr=snr,
            dr=0.075,
            use_existing_kdp=False,
            clear_air_mode="mask",
        )

        self.assertTrue(labeled["clear_air_mask"][0, 0])
        self.assertTrue(labeled["qc_mask"][1, 1])
        self.assertFalse(masked["qc_mask"][0, 0])
        self.assertTrue(np.isnan(masked["ref_corrected"][0, 0]))

    def test_apply_dualpol_qc_adds_fields_without_mutating_original_when_copying(self):
        from pycwr.qc import apply_dualpol_qc

        prd = self._build_synthetic_prd()
        processed = apply_dualpol_qc(prd, inplace=False)

        self.assertNotIn("Zc", prd.fields[0].data_vars)
        for field_name in (
            "Zc",
            "ZDRc",
            "PhiDP_smooth",
            "KDPc",
            "PhiDP_texture",
            "METEO_MASK",
            "CLEAR_AIR_MASK",
            "QC_MASK",
            "PIA",
            "PIA_ZDR",
        ):
            self.assertIn(field_name, processed.fields[0].data_vars)

    def test_apply_dualpol_qc_inplace_mutates_target(self):
        from pycwr.qc import apply_dualpol_qc

        prd = self._build_synthetic_prd()
        returned = apply_dualpol_qc(prd, inplace=True)

        self.assertIs(returned, prd)
        self.assertIn("Zc", prd.fields[0].data_vars)

    def test_prd_apply_dualpol_qc_method_writes_results(self):
        prd = self._build_synthetic_prd()

        processed = prd.apply_dualpol_qc(inplace=False)

        self.assertIn("Zc", processed.fields[0].data_vars)
        self.assertIn("QC_MASK", processed.fields[0].data_vars)
        self.assertNotIn("Zc", prd.fields[0].data_vars)

    def test_qc_fields_export_to_pyart(self):
        prd = self._build_synthetic_prd()
        processed = prd.apply_dualpol_qc(inplace=False)

        radar = processed.to_pyart_radar(use_external=False, force_rebuild=True)

        self.assertIn("corrected_reflectivity", radar.fields)
        self.assertIn("corrected_differential_reflectivity", radar.fields)

    def test_apply_dualpol_qc_requires_dualpol_phase_information(self):
        from pycwr.qc import apply_dualpol_qc

        prd = self._build_synthetic_prd()
        prd.fields[0] = prd.fields[0].drop_vars(["KDP", "PhiDP"])

        with self.assertRaises(ValueError):
            apply_dualpol_qc(prd, inplace=False)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import numpy as np
from .BaseDataProtocol.SABProtocol import dtype_sab
from .util import _prepare_for_read, _read_all, _unpack_from_buf, _validate_offset_length, julian2date, \
    get_radar_info, make_time_unit_str, get_radar_sitename, date2num
from ..core.NRadar import PRD
from ..configure.pyart_config import get_metadata, get_fillvalue
from ..configure.default_config import CINRAD_field_mapping, _LIGHT_SPEED
from ..core.PyartRadar import Radar


class SABBaseData(object):
    """Decode SA/SB/CB/SC2.0 base data into a lightweight radial structure."""

    def __init__(self, filename, station_lon=None, station_lat=None, station_alt=None):
        """
        :param filename:  radar basedata filename
        :param station_lon:  radar station longitude //units: degree east
        :param station_lat:  radar station latitude //units:degree north
        :param station_alt:  radar station altitude //units: meters
        """
        super(SABBaseData, self).__init__()
        self.filename = filename
        self.station_lon = station_lon
        self.station_lat = station_lat
        self.station_alt = station_alt
        self.fid = _prepare_for_read(self.filename)
        self._raw_buf = _read_all(self.fid, "SAB radial payload")
        self.fid.close()
        self.RadialNum, self.nrays = self._determine_radial_record_layout()
        self.radial, self._status, self._azimuth, self._elevation, self._julian_date, self._msends, \
            self._nyquist, self._unambiguous_range = self._parse_radial()
        self._scan_time = None
        self.sweep_start_ray_index = np.where((self._status == 0) | (self._status == 3))[0]
        self.sweep_end_ray_index = np.where((self._status == 2) | (self._status == 4))[0]
        self.nsweeps = len(self.sweep_start_ray_index)
        self._raw_buf = None

    def _determine_radial_record_layout(self):
        """Determine the radial record size and radar family from the raw buffer."""
        if len(self._raw_buf) < 16:
            raise ValueError("SAB file header is truncated.")
        if self._raw_buf[14:16] != b'\x01\x00':
            raise ValueError("File is not a valid SA/SB/CB file.")
        data_len = len(self._raw_buf)
        if (data_len % 2432 != 0) and (data_len % 4132 != 0) and (data_len % 3132 != 0):
            raise ValueError("SAB file size is inconsistent with the supported radial layouts.")
        # Distinguish SA/SB, CB, and SC layouts by fixed radial record size.
        if data_len % 2432 == 0:
            RadialNum = 2432
            self.Type = "SAB"
        elif data_len % 4132 == 0:
            RadialNum = 4132
            self.Type = 'CB'
        else:
            RadialNum = 3132
            self.Type = 'SC'
        return RadialNum, int(data_len / RadialNum)

    def _RadialNum_SAB_CB(self):
        """Backward-compatible alias for ``_determine_radial_record_layout``."""
        return self._determine_radial_record_layout()

    def _parse_radial(self):
        """Parse all radial records from the in-memory SAB buffer."""
        radial = []
        status = []
        azimuth = []
        elevation = []
        julian_date = []
        msends = []
        nyquist = []
        unambiguous_range = []
        buf = self._raw_buf
        for pos in range(0, len(buf), self.RadialNum):
            iray = self._parse_radial_single(buf[pos:pos + self.RadialNum])
            radial.append(iray)
            status.append(iray['RadialStatus'])
            azimuth.append(iray['AZ'] / 8. * 180. / 4096.)
            elevation.append(iray['El'] / 8. * 180. / 4096.)
            julian_date.append(iray['JulianDate'])
            msends.append(iray['mSends'])
            nyquist.append(iray['Nyquist'] / 100.)
            unambiguous_range.append(iray['URange'] / 10.)
        return radial, np.asarray(status), np.asarray(azimuth), np.asarray(elevation), \
            np.asarray(julian_date), np.asarray(msends), np.asarray(nyquist), np.asarray(unambiguous_range)

    def _parse_radial_single(self, radial_buf):
        Radial = {}
        RadialHeader, size_tmp = _unpack_from_buf(radial_buf, 0, dtype_sab.RadialHeader())
        Radial.update(RadialHeader)
        ref_offset, ref_count = _validate_offset_length(
            len(radial_buf),
            int(RadialHeader['PtrOfReflectivity']) + dtype_sab.InfSize,
            int(RadialHeader['GatesNumberOfReflectivity']),
            "SAB reflectivity block",
        )
        dop_offset, dop_count = _validate_offset_length(
            len(radial_buf),
            int(RadialHeader['PtrOfVelocity']) + dtype_sab.InfSize,
            int(RadialHeader['GatesNumberOfDoppler']),
            "SAB velocity block",
        )
        sw_offset, sw_count = _validate_offset_length(
            len(radial_buf),
            int(RadialHeader['PtrOfSpectrumWidth']) + dtype_sab.InfSize,
            int(RadialHeader['GatesNumberOfDoppler']),
            "SAB spectrum-width block",
        )
        dBZ = np.frombuffer(
            radial_buf,
            dtype=np.uint8,
            count=ref_count,
            offset=ref_offset,
        )
        V = np.frombuffer(
            radial_buf,
            dtype=np.uint8,
            count=dop_count,
            offset=dop_offset,
        )
        W = np.frombuffer(
            radial_buf,
            dtype=np.uint8,
            count=sw_count,
            offset=sw_offset,
        )
        Radial['fields'] = {}
        Radial['fields']['dBZ'] = self._decode_field(dBZ, -32.0)
        Radial['fields']['V'] = self._decode_field(V, -63.5)
        Radial['fields']['W'] = self._decode_field(W, -63.5)
        return Radial

    @staticmethod
    def _decode_field(raw, base):
        out = np.full(raw.shape, np.nan, dtype=np.float32)
        valid = raw > 1
        if np.any(valid):
            out[valid] = (raw[valid].astype(np.float32) - 2.0) / 2.0 + base
        return out

    def get_nyquist_velocity(self):
        """Return the per-ray Nyquist velocity."""
        return self._nyquist

    def get_unambiguous_range(self):
        """Return the per-ray unambiguous range in kilometers."""
        return self._unambiguous_range

    def get_scan_time(self):
        """Return the acquisition time for each ray."""
        if self._scan_time is None:
            self._scan_time = np.array(
                [julian2date(julian, ms) for julian, ms in zip(self._julian_date, self._msends)]
            )
        return self._scan_time

    def get_sweep_end_ray_index(self):
        """Return the inclusive end index of each sweep."""
        return self.sweep_end_ray_index

    def get_sweep_start_ray_index(self):
        """Return the start index of each sweep."""
        return self.sweep_start_ray_index

    def get_rays_per_sweep(self):
        """Return the number of rays in each sweep."""
        return self.sweep_end_ray_index - self.sweep_start_ray_index + 1

    def get_azimuth(self):
        """Return the azimuth angle for each ray."""
        return self._azimuth

    def get_elevation(self):
        """Return the elevation angle for each ray."""
        return self._elevation

    def get_latitude_longitude_altitude_frequency(self):
        """Return latitude, longitude, altitude, and radar frequency."""
        lat, lon, alt, frequency = get_radar_info(self.filename)
        if self.station_lon is not None:
            lon = self.station_lon
        if self.station_lat is not None:
            lat = self.station_lat
        if self.station_alt is not None:
            alt = self.station_alt
        return lat, lon, alt, frequency

    def get_scan_type(self):
        """Return the scan type string."""
        return "ppi"

    def get_sitename(self):
        return get_radar_sitename(self.filename)


class SAB2NRadar(object):
    """Bridge from raw SAB data to an NRadar object."""

    def __init__(self, SAB):
        self.SAB = SAB
        self._dbz_index_cache = {}
        self.v_index_alone = self.get_v_idx()
        self.dBZ_index_alone = self.get_dbz_idx()
        self.dBZ_Res = self.SAB.radial[0]["GateSizeOfReflectivity"]
        sab_elevation = self.SAB.get_elevation()
        for index_with_dbz, index_with_v in zip(self.dBZ_index_alone, self.v_index_alone):
            assert abs(sab_elevation[index_with_v] - sab_elevation[index_with_dbz]) < 0.5, "warning! maybe it is a problem."
            self.interp_dBZ(index_with_dbz, index_with_v)
        keep_mask = np.ones(self.SAB.nrays, dtype=bool)
        keep_mask[self.get_remove_radial_indices()] = False
        self.radial = [iray for iray, keep in zip(self.SAB.radial, keep_mask) if keep]
        self._azimuth = self.SAB.get_azimuth()[keep_mask]
        self._elevation = self.SAB.get_elevation()[keep_mask]
        self._julian_date = self.SAB._julian_date[keep_mask]
        self._msends = self.SAB._msends[keep_mask]
        self._nyquist = self.SAB.get_nyquist_velocity()[keep_mask]
        self._unambiguous_range = self.SAB.get_unambiguous_range()[keep_mask]
        self._scan_time = None
        self.nrays = len(self.radial)
        self.nsweeps = self.SAB.nsweeps - self.dBZ_index_alone.size
        status = self.SAB._status[keep_mask]
        self.sweep_start_ray_index = np.where((status == 0) | (status == 3))[0]
        self.sweep_end_ray_index = np.where((status == 2) | (status == 4))[0]
        self.scan_type = self.SAB.get_scan_type()
        self.latitude, self.longitude, self.altitude, self.frequency = \
            self.SAB.get_latitude_longitude_altitude_frequency()
        self.bins_per_sweep = self.get_nbins_per_sweep()
        self.max_bins = self.bins_per_sweep.max()
        self.range = self.get_range_per_radial(self.max_bins)
        self.azimuth = self.get_azimuth()
        self.elevation = self.get_elevation()
        self.extended_fields = self._build_extended_fields()
        self.fields = self._get_fields()
        self.sitename = self.SAB.get_sitename()

    def get_remove_radial_indices(self):
        """Return the radial indices removed after sweep alignment."""
        index_romove = []
        for isweep in self.dBZ_index_alone:
            index_romove.extend(range(self.SAB.sweep_start_ray_index[isweep], \
                                      self.SAB.sweep_end_ray_index[isweep] + 1))
        return index_romove

    def get_reomve_radial_num(self):
        """Backward-compatible alias for ``get_remove_radial_indices``."""
        return self.get_remove_radial_indices()

    def get_v_idx(self):
        """Return sweep indices that contain Doppler moments without reflectivity."""
        flag = np.array([((self.SAB.radial[idx]['fields']["V"].size != 0) and
                          (self.SAB.radial[idx]['fields']["dBZ"].size == 0)) \
                         for idx in self.SAB.sweep_start_ray_index])
        return np.where(flag == 1)[0]

    def get_dbz_idx(self):
        """Return sweep indices that contain reflectivity without Doppler moments."""
        flag = np.array([((self.SAB.radial[idx]['fields']["V"].size == 0) and
                          (self.SAB.radial[idx]['fields']["dBZ"].size != 0)) \
                         for idx in self.SAB.sweep_start_ray_index])
        return np.where(flag == 1)[0]

    def interp_dBZ(self, field_with_dBZ_num, field_without_dBZ_num):
        """
        Copy reflectivity rays onto the matching Doppler-only sweep by azimuth.
        :param field_with_dBZ_num: source sweep index, 0-based.
        :param field_without_dBZ_num: target sweep index, 0-based.
        """
        azimuth = self.SAB.get_azimuth()
        assert (field_with_dBZ_num + 1) == field_without_dBZ_num, "check interp sweep!"
        dbz_az = azimuth[self.SAB.sweep_start_ray_index[field_with_dBZ_num]: \
                         self.SAB.sweep_end_ray_index[field_with_dBZ_num] + 1]
        v_az = azimuth[self.SAB.sweep_start_ray_index[field_without_dBZ_num]: \
                       self.SAB.sweep_end_ray_index[field_without_dBZ_num] + 1]
        dbz_idx = np.argmin(np.abs(dbz_az.reshape(-1, 1) - v_az.reshape(1, -1)), axis=0) + \
                  self.SAB.sweep_start_ray_index[field_with_dBZ_num]
        v_idx = np.arange(self.SAB.sweep_start_ray_index[field_without_dBZ_num], \
                          self.SAB.sweep_end_ray_index[field_without_dBZ_num] + 1)
        for ind_dbz, ind_v in zip(dbz_idx, v_idx):
            self.SAB.radial[ind_v]["fields"]['dBZ'] = self.SAB.radial[ind_dbz]["fields"]['dBZ']

    def get_azimuth(self):
        """Return the azimuth angle for each retained ray."""
        return self._azimuth

    def get_elevation(self):
        """Return the elevation angle for each retained ray."""
        return self._elevation

    def get_rays_per_sweep(self):
        """Return the number of rays in each retained sweep."""
        return self.sweep_end_ray_index - self.sweep_start_ray_index + 1

    def get_scan_time(self):
        """Return the acquisition time for each retained ray."""
        if self._scan_time is None:
            self._scan_time = np.array(
                [julian2date(julian, ms) for julian, ms in zip(self._julian_date, self._msends)]
            )
        return self._scan_time

    def get_nyquist_velocity(self):
        """Return the per-ray Nyquist velocity."""
        return self._nyquist

    def get_unambiguous_range(self):
        """Return the per-ray unambiguous range."""
        return self._unambiguous_range

    def get_sweep_end_ray_index(self):
        """Return the inclusive end index of each retained sweep."""
        return self.sweep_end_ray_index

    def get_sweep_start_ray_index(self):
        """Return the start index of each retained sweep."""
        return self.sweep_start_ray_index

    def get_nbins_per_sweep(self):
        """Return the Doppler gate count for each retained sweep."""
        return np.array([self.radial[idx]['fields']['V'].size for idx in self.sweep_start_ray_index])

    def get_range_per_radial(self, length):
        """Return Doppler gate-center ranges for a radial of ``length`` bins."""
        Resolution = self.radial[0]["GateSizeOfDoppler"]
        return np.linspace(Resolution, Resolution * length, length)

    def get_dbz_range_per_radial(self, length):
        """Return reflectivity gate-center ranges for a radial of ``length`` bins."""
        Resolution = self.dBZ_Res
        start_range = self.radial[0]["GateSizeOfDoppler"]
        return np.linspace(start_range, start_range + Resolution * (length - 1), length)

    def _get_fields(self):
        """Assemble the retained fields into dense 2-D arrays."""
        field_keys = tuple(self.radial[0]['fields'].keys())
        fields = {ikey: np.full((self.nrays, self.max_bins), np.nan, dtype=np.float64) for ikey in field_keys}
        for iray, radial in enumerate(self.radial):
            radial_fields = radial['fields']
            for ikey in field_keys:
                fields[ikey][iray, :] = self._add_or_del_field(radial_fields.get(ikey), ikey)
        return fields

    def _build_extended_fields(self):
        """Collect native-range reflectivity for sweeps longer than the aligned grid."""
        extended_sweeps = {}
        scan_time = self.get_scan_time()
        for sweep_idx, (start, end, aligned_bins) in enumerate(
            zip(self.sweep_start_ray_index, self.sweep_end_ray_index, self.bins_per_sweep)
        ):
            native_bins = max(
                (
                    self.radial[iray]["fields"].get("dBZ", np.empty((0,), dtype=np.float32)).size
                    for iray in range(start, end + 1)
                ),
                default=0,
            )
            if native_bins <= int(aligned_bins):
                continue

            data = np.full((end - start + 1, native_bins), np.nan, dtype=np.float32)
            for out_idx, iray in enumerate(range(start, end + 1)):
                dat_ray = self.radial[iray]["fields"].get("dBZ")
                if dat_ray is None or dat_ray.size == 0:
                    continue
                copy_len = min(dat_ray.size, native_bins)
                data[out_idx, :copy_len] = dat_ray[:copy_len].astype(np.float32, copy=False)

            extended_sweeps[sweep_idx] = {
                "data": data,
                "range": self.get_dbz_range_per_radial(native_bins).astype(np.float32, copy=False),
                "time": np.asarray(scan_time[start:end + 1]),
                "azimuth": np.asarray(self._azimuth[start:end + 1], dtype=np.float32),
                "elevation": np.asarray(self._elevation[start:end + 1], dtype=np.float32),
                "aligned_bins": int(aligned_bins),
            }

        return {"dBZ": extended_sweeps} if extended_sweeps else {}

    def _get_dbz_resample_index(self, source_length):
        cached = self._dbz_index_cache.get(source_length)
        if cached is not None:
            return cached

        source = self.get_dbz_range_per_radial(source_length)
        target = self.range
        right = np.searchsorted(source, target, side="left")
        left = np.clip(right - 1, 0, source_length - 1)
        right = np.clip(right, 0, source_length - 1)
        choose_right = np.abs(source[right] - target) < np.abs(target - source[left])
        nearest = np.where(choose_right, right, left)
        valid = (target >= source[0]) & (target <= source[-1])
        self._dbz_index_cache[source_length] = (valid, nearest)
        return valid, nearest

    def _add_or_del_field(self, dat_ray, key):
        """
        Normalize a radial field to the common Doppler-aligned range grid.
        :param dat_ray: radial field data.
        :param key: field name.
        """
        length = self.max_bins
        if key == "dBZ":
            valid, nearest = self._get_dbz_resample_index(dat_ray.size)
            out = np.full((length,), np.nan)
            out[valid] = dat_ray[nearest[valid]]
            return out
        if dat_ray.size >= length:
            return (dat_ray[:length]).ravel()
        else:
            out = np.full((length,), np.nan)
            out[:dat_ray.size] = dat_ray
            return out.ravel()

    def get_nradar_nyquist_speed(self):
        """array shape (nsweeps)"""
        return np.array([self.radial[idx]['Nyquist'] / 100. for idx in self.sweep_start_ray_index])

    def get_NRadar_nyquist_speed(self):
        """Backward-compatible alias for ``get_nradar_nyquist_speed``."""
        return self.get_nradar_nyquist_speed()

    def get_nradar_unambiguous_range(self):
        """array shape (nsweeps)"""
        return np.array([self.radial[idx]['URange'] / 10. for idx in self.sweep_start_ray_index])

    def get_NRadar_unambiguous_range(self):
        """Backward-compatible alias for ``get_nradar_unambiguous_range``."""
        return self.get_nradar_unambiguous_range()

    def get_fixed_angle(self):
        if self.nsweeps == 9:
            fixed_angle = np.array([0.50, 1.45, 2.40, 3.35, 4.30, 6.00, 9.00, 14.6, 19.5])
        elif self.nsweeps == 14:
            fixed_angle = np.array([0.50, 1.45, 2.40, 3.35, 4.30, 5.25, 6.2, 7.5, 8.7, 10, 12, 14, 16.7, 19.5])
        elif self.nsweeps == 6:
            fixed_angle = np.array([0.50, 1.50, 2.50, 2.50, 3.50, 4.50])
        elif self.nsweeps == 4:
            fixed_angle = np.array([0.50, 2.50, 3.50, 4.50])
        else:
            fixed_angle = np.array([self.radial[idx]['El'] / 8. * 180. / 4096. for idx in self.sweep_start_ray_index])
        return fixed_angle

    def to_prd(self, effective_earth_radius=None):
        """Build an ``NRadar.PRD`` object from the decoded SAB volume."""
        return PRD(fields=self.fields, scan_type=self.scan_type, time=self.get_scan_time(), \
                          range=self.range, azimuth=self.azimuth, elevation=self.elevation, latitude=self.latitude, \
                          longitude=self.longitude, altitude=self.altitude,
                          sweep_start_ray_index=self.sweep_start_ray_index, \
                          sweep_end_ray_index=self.sweep_end_ray_index, fixed_angle=self.get_fixed_angle(), \
                          bins_per_sweep=self.bins_per_sweep, nyquist_velocity=self.get_nradar_nyquist_speed(), \
                          frequency=self.frequency, unambiguous_range=self.get_nradar_unambiguous_range(), \
                          nrays=self.nrays, nsweeps=self.nsweeps, sitename = self.sitename,
                          pyart_radar=None, effective_earth_radius=effective_earth_radius,
                          extended_fields=self.extended_fields,
                          metadata={"original_container": "CINRAD/SAB", "radar_name": "CINRAD/SA/SB/CB/SC"})

    def ToPRD(self, effective_earth_radius=None):
        """Backward-compatible alias for ``to_prd``."""
        return self.to_prd(effective_earth_radius=effective_earth_radius)

    def to_pyart_radar(self, effective_earth_radius=None, **kwargs):
        """Export the decoded SAB volume through the PRD Py-ART adapter."""
        return self.to_prd(effective_earth_radius=effective_earth_radius).to_pyart_radar(**kwargs)

    def ToPyartRadar(self, effective_earth_radius=None, **kwargs):
        """Backward-compatible alias for ``to_pyart_radar``."""
        return self.to_pyart_radar(effective_earth_radius=effective_earth_radius, **kwargs)

    def _get_instrument_parameters(self):
        """ Return a dictionary containing instrument parameters. """

        # pulse width
        pulse_width = get_metadata('pulse_width')
        pulse_width['data'] = np.array([self.radial[0]["GateSizeOfDoppler"] / _LIGHT_SPEED,], dtype='float32')  # m->sec

        # assume that the parameters in the first ray represent the beam widths,
        # bandwidth and frequency in the entire volume

        wavelength_hz = self.frequency * 10 ** 9

        # radar_beam_width_h
        radar_beam_width_h = get_metadata('radar_beam_width_h')
        radar_beam_width_h['data'] = np.array([1, ], dtype='float32')

        # radar_beam_width_v
        radar_beam_width_v = get_metadata('radar_beam_width_v')
        radar_beam_width_v['data'] = np.array([1, ], dtype='float32')

        # frequency
        frequency = get_metadata('frequency')
        frequency['data'] = np.array([wavelength_hz], dtype='float32')

        instrument_parameters = {
            'pulse_width': pulse_width,
            'radar_beam_width_h': radar_beam_width_h,
            'radar_beam_width_v': radar_beam_width_v,
            'frequency': frequency, }

        # nyquist velocity if defined
        nyquist_velocity = get_metadata('nyquist_velocity')
        nyquist_velocity['data'] = self.get_nyquist_velocity()
        instrument_parameters['nyquist_velocity'] = nyquist_velocity
        return instrument_parameters



"""Section and RHI extraction examples plus regression coverage."""

import unittest

import matplotlib
import numpy as np

matplotlib.use("Agg")
import matplotlib.pyplot as plt


class SectionExtractionTests(unittest.TestCase):
    def _build_ppi_prd(self, azimuths=None, ranges=None, elevations=(1.0, 3.0)):
        from pycwr.core.NRadar import PRD

        azimuths = np.asarray(azimuths if azimuths is not None else np.arange(0.0, 360.0, 45.0), dtype=np.float64)
        ranges = np.asarray(ranges if ranges is not None else np.arange(1000.0, 9000.0, 1000.0), dtype=np.float64)
        nsweeps = len(elevations)
        nrays_per_sweep = azimuths.size
        nrays = nsweeps * nrays_per_sweep
        azimuth = np.tile(azimuths, nsweeps)
        elevation = np.repeat(np.asarray(elevations, dtype=np.float64), nrays_per_sweep)
        base_field = azimuth[:, None] + ranges[None, :] / 1000.0
        fields = {
            "dBZ": base_field.copy(),
            "ZDR": base_field.copy() * 0.1,
        }
        sweep_start = np.arange(nsweeps, dtype=np.int32) * nrays_per_sweep
        sweep_end = sweep_start + nrays_per_sweep - 1
        return PRD(
            fields=fields,
            scan_type="ppi",
            time=np.arange(nrays),
            range=ranges,
            azimuth=azimuth,
            elevation=elevation,
            latitude=30.0,
            longitude=120.0,
            altitude=100.0,
            sweep_start_ray_index=sweep_start,
            sweep_end_ray_index=sweep_end,
            fixed_angle=np.asarray(elevations, dtype=np.float64),
            bins_per_sweep=np.full(nsweeps, ranges.size, dtype=np.int32),
            nyquist_velocity=np.full(nsweeps, 15.0, dtype=np.float64),
            frequency=5.6,
            unambiguous_range=np.full(nsweeps, 150000.0, dtype=np.float64),
            nrays=nrays,
            nsweeps=nsweeps,
            sitename="TEST",
        )

    def _build_rhi_prd(self):
        from pycwr.core.NRadar import PRD

        ranges = np.arange(1000.0, 5000.0, 1000.0)
        elevation = np.array([0.5, 1.5, 2.5], dtype=np.float64)
        azimuth = np.full(elevation.shape, 45.0, dtype=np.float64)
        fields = {
            "dBZ": elevation[:, None] + ranges[None, :] / 1000.0,
        }
        return PRD(
            fields=fields,
            scan_type="rhi",
            time=np.arange(elevation.size),
            range=ranges,
            azimuth=azimuth,
            elevation=elevation,
            latitude=30.0,
            longitude=120.0,
            altitude=100.0,
            sweep_start_ray_index=np.array([0], dtype=np.int32),
            sweep_end_ray_index=np.array([elevation.size - 1], dtype=np.int32),
            fixed_angle=np.array([45.0], dtype=np.float64),
            bins_per_sweep=np.array([ranges.size], dtype=np.int32),
            nyquist_velocity=np.array([15.0], dtype=np.float64),
            frequency=5.6,
            unambiguous_range=np.array([150000.0], dtype=np.float64),
            nrays=elevation.size,
            nsweeps=1,
            sitename="RHI",
        )

    def test_extract_section_matches_linear_analytic_field(self):
        from pycwr.core.transforms import cartesian_to_antenna_cwr

        prd = self._build_ppi_prd()
        section = prd.extract_section(
            (1000.0, 2000.0),
            (5000.0, 2000.0),
            field_name="dBZ",
            point_units="m",
        )

        self.assertEqual(section["dBZ"].shape[0], 2)
        self.assertEqual(section.attrs["interpolation"], "linear")
        for strip_index, elevation in enumerate((1.0, 3.0)):
            azimuth, ranges, _ = cartesian_to_antenna_cwr(
                section["x"].isel(sweep=strip_index).values,
                section["y"].isel(sweep=strip_index).values,
                elevation,
                float(prd.scan_info.altitude.values),
                effective_earth_radius=prd.effective_earth_radius,
            )
            expected = azimuth + ranges / 1000.0
            np.testing.assert_allclose(
                section["dBZ"].isel(sweep=strip_index).values,
                expected,
                atol=1e-6,
                equal_nan=True,
            )

    def test_extract_rhi_handles_azimuth_wraparound(self):
        prd = self._build_ppi_prd(azimuths=np.array([5.0, 90.0, 180.0, 355.0]), elevations=(1.0,))
        sweep = prd.fields[0]
        custom = np.vstack(
            [
                np.full(sweep.range.size, 20.0),
                np.full(sweep.range.size, 40.0),
                np.full(sweep.range.size, 60.0),
                np.full(sweep.range.size, 10.0),
            ]
        )
        prd.fields[0]["dBZ"] = (("time", "range"), custom)
        section = prd.extract_rhi(azimuth=359.0, field_name="dBZ")

        np.testing.assert_allclose(section["dBZ"].isel(sweep=0).values, np.full(sweep.range.size, 14.0))

    def test_extract_rhi_supports_native_rhi_scan(self):
        prd = self._build_rhi_prd()
        section = prd.extract_rhi(field_name="dBZ")

        self.assertEqual(section.attrs["scan_type"], "rhi")
        self.assertEqual(section["dBZ"].shape, (3, 4))
        np.testing.assert_allclose(section["dBZ"].values, prd.fields[0]["dBZ"].values)
        np.testing.assert_array_equal(section["source_sweep"].values, np.zeros(3, dtype=np.int32))

    def test_legacy_section_helpers_remain_compatible(self):
        prd = self._build_ppi_prd()

        mesh_xy, mesh_z, field_data = prd.get_vcs_data((1000.0, 0.0), (6000.0, 0.0), "dBZ")
        rhi_xy, rhi_z, rhi_field = prd.get_RHI_data(45.0, "dBZ")

        self.assertEqual(len(mesh_xy), prd.nsweeps)
        self.assertEqual(len(mesh_z), prd.nsweeps)
        self.assertEqual(len(field_data), prd.nsweeps)
        self.assertEqual(len(rhi_xy), prd.nsweeps)
        self.assertEqual(rhi_xy[0].shape, rhi_z[0].shape)
        self.assertEqual(field_data[0].shape[0], 1)
        self.assertEqual(rhi_field[0].shape[0], 1)


class SectionPlotTests(unittest.TestCase):
    def _build_plot_prd(self):
        return SectionExtractionTests()._build_ppi_prd()

    def test_easy_plot_section_smoke(self):
        from pycwr.draw.easy import plot_section

        prd = self._build_plot_prd()
        result = plot_section(
            prd,
            start=(1.0, 2.0),
            end=(5.0, 2.0),
            field="dBZ",
            show=False,
        )
        try:
            self.assertIsNotNone(result.artist)
        finally:
            plt.close(result.fig)

    def test_easy_plot_rhi_smoke(self):
        from pycwr.draw.easy import plot_rhi

        prd = self._build_plot_prd()
        result = plot_rhi(
            prd,
            azimuth=45.0,
            field="dBZ",
            show=False,
        )
        try:
            self.assertIsNotNone(result.artist)
        finally:
            plt.close(result.fig)


if __name__ == "__main__":
    unittest.main()



from importlib import import_module

__all__ = [
    "apply_dualpol_qc",
    "build_clear_air_mask",
    "build_meteo_mask",
    "correct_attenuation",
    "correct_attenuation_HB",
    "correct_attenuation_kdp",
    "despeckle_mask",
    "kdp_from_phidp",
    "phidp_texture",
    "pia_from_kdp",
    "QC_REFERENCE_NOTES",
    "resolve_kdp_coefficients",
    "run_dualpol_qc",
    "smooth_phidp",
]

_LAZY_SYMBOLS = {
    "correct_attenuation": ("attenuation", "correct_attenuation"),
    "correct_attenuation_HB": ("attenuation", "correct_attenuation_HB"),
    "correct_attenuation_kdp": ("attenuation", "correct_attenuation_kdp"),
    "pia_from_kdp": ("attenuation", "pia_from_kdp"),
    "resolve_kdp_coefficients": ("attenuation", "resolve_kdp_coefficients"),
    "apply_dualpol_qc": ("pipeline", "apply_dualpol_qc"),
    "run_dualpol_qc": ("pipeline", "run_dualpol_qc"),
    "QC_REFERENCE_NOTES": ("polarimetric", "QC_REFERENCE_NOTES"),
    "build_clear_air_mask": ("polarimetric", "build_clear_air_mask"),
    "build_meteo_mask": ("polarimetric", "build_meteo_mask"),
    "despeckle_mask": ("polarimetric", "despeckle_mask"),
    "kdp_from_phidp": ("polarimetric", "kdp_from_phidp"),
    "phidp_texture": ("polarimetric", "phidp_texture"),
    "smooth_phidp": ("polarimetric", "smooth_phidp"),
}


def __getattr__(name):
    symbol = _LAZY_SYMBOLS.get(name)
    if symbol is not None:
        module = import_module(f".{symbol[0]}", __name__)
        value = getattr(module, symbol[1])
        globals()[name] = value
        return value
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from importlib import import_module

__all__ = [
    "WIND_REFERENCE_NOTES",
    "VAD",
    "VVP",
    "vad",
    "vvp",
    "retrieve_vad",
    "retrieve_vvp",
    "retrieve_vwp",
    "select_velocity_field",
    "fhc_HCL",
    "fhc_hcl",
    "hid_beta_function",
    "classify_hydrometeors",
    "classify_sweep_hydrometeors",
    "apply_hydrometeor_classification",
    "interpolate_temperature_profile",
    "available_hydrometeor_classes",
    "hydrometeor_class_name",
    "HYDROMETEOR_CLASSES",
    "HYDROMETEOR_CLASS_IDS",
    "HID_REFERENCE_NOTES",
]


def __getattr__(name):
    if name in {"VAD", "VVP", "vad", "vvp", "retrieve_vad", "retrieve_vvp", "retrieve_vwp", "select_velocity_field", "WIND_REFERENCE_NOTES"}:
        module = import_module(".WindField", __name__)
        value = getattr(module, name)
        globals()[name] = value
        return value
    if name in {
        "fhc_HCL",
        "fhc_hcl",
        "hid_beta_function",
        "classify_hydrometeors",
        "classify_sweep_hydrometeors",
        "apply_hydrometeor_classification",
        "interpolate_temperature_profile",
        "available_hydrometeor_classes",
        "hydrometeor_class_name",
        "HYDROMETEOR_CLASSES",
        "HYDROMETEOR_CLASS_IDS",
        "HID_REFERENCE_NOTES",
    }:
        module = import_module(".HID", __name__)
        value = getattr(module, name)
        globals()[name] = value
        return value
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# -*- coding: utf-8 -*-
import datetime
import os
import struct
from collections import OrderedDict

import numpy as np
from .BaseDataProtocol.WSR98DProtocol import dtype_98D
from .util import (
    _prepare_for_read,
    _read_all,
    _read_exact,
    _unpack_from_buf,
    _validate_count,
    julian2date_SEC,
    make_time_unit_str,
    date2num,
)
from ..core.NRadar import PRD
from ..configure.pyart_config import get_metadata, get_fillvalue
from ..configure.default_config import CINRAD_field_mapping
from ..core.PyartRadar import Radar


MAX_WSR98D_SWEEPS = 512
MAX_WSR98D_MOMENTS_PER_RADIAL = 64
MAX_WSR98D_MOMENT_DATA_BYTES = 32 * 1024 * 1024
WSR98D_MAX_DECODED_PAYLOAD_BYTES = 1024 * 1024 * 1024

WSR98D_WRITE_FIELD_SPECS = OrderedDict(
    [
        ("dBT", {"data_type": 1, "scale": 2, "offset": 66, "bin_length": 1}),
        ("dBZ", {"data_type": 2, "scale": 2, "offset": 66, "bin_length": 1}),
        ("V", {"data_type": 3, "scale": 2, "offset": 129, "bin_length": 1}),
        ("W", {"data_type": 4, "scale": 2, "offset": 129, "bin_length": 1}),
        ("SQI", {"data_type": 5, "scale": 200, "offset": 0, "bin_length": 1}),
        ("CPA", {"data_type": 6, "scale": 2, "offset": 129, "bin_length": 1}),
        ("ZDR", {"data_type": 7, "scale": 16, "offset": 130, "bin_length": 1}),
        ("LDR", {"data_type": 8, "scale": 16, "offset": 130, "bin_length": 1}),
        ("CC", {"data_type": 9, "scale": 200, "offset": 5, "bin_length": 1}),
        ("PhiDP", {"data_type": 10, "scale": 100, "offset": 50, "bin_length": 2}),
        ("KDP", {"data_type": 11, "scale": 10, "offset": 50, "bin_length": 1}),
        ("CP", {"data_type": 12, "scale": 2, "offset": 129, "bin_length": 1}),
        ("HCL", {"data_type": 14, "scale": 1, "offset": 0, "bin_length": 1}),
        ("CF", {"data_type": 15, "scale": 2, "offset": 129, "bin_length": 1}),
        ("SNRH", {"data_type": 16, "scale": 2, "offset": 20, "bin_length": 1}),
        ("SNRV", {"data_type": 17, "scale": 2, "offset": 20, "bin_length": 1}),
        ("Zc", {"data_type": 32, "scale": 2, "offset": 66, "bin_length": 1}),
        ("Vc", {"data_type": 33, "scale": 2, "offset": 129, "bin_length": 1}),
        ("Wc", {"data_type": 34, "scale": 2, "offset": 129, "bin_length": 1}),
        ("ZDRc", {"data_type": 35, "scale": 16, "offset": 130, "bin_length": 1}),
    ]
)

_WSR98D_RADIAL_HEADER = dtype_98D.RadialHeader()
_WSR98D_RADIAL_HEADER_KEYS = tuple(field_name for field_name, _ in _WSR98D_RADIAL_HEADER)
_WSR98D_RADIAL_HEADER_STRUCT = struct.Struct("<" + "".join(field_type for _, field_type in _WSR98D_RADIAL_HEADER))
_WSR98D_MOMENT_HEADER = dtype_98D.RadialData()
_WSR98D_MOMENT_HEADER_KEYS = tuple(field_name for field_name, _ in _WSR98D_MOMENT_HEADER)
_WSR98D_MOMENT_HEADER_STRUCT = struct.Struct("<" + "".join(field_type for _, field_type in _WSR98D_MOMENT_HEADER))


def _decode_wsr98d_resolution(raw_value):
    """Decode meter resolution fields, including the vendor 62.5 m fixed-point encoding."""
    raw_value = int(raw_value)
    if raw_value >= 32768:
        return (raw_value - 32768) / 100.0
    return float(raw_value)


def _encode_wsr98d_resolution(resolution_m):
    """Encode a meter resolution back into the on-disk WSR98D integer representation."""
    resolution_m = float(resolution_m)
    rounded_meter = round(resolution_m)
    if abs(resolution_m - rounded_meter) < 1.0e-6:
        return int(rounded_meter)
    return int(round(resolution_m * 100.0 + 32768.0))


def _ensure_ppi_prd_volume(prd):
    scan_type = str(np.asarray(prd.scan_info["scan_type"].values).item())
    if scan_type != "ppi":
        raise ValueError("WSR98D export only supports ppi PRD volumes.")


def _ensure_output_path(filename, overwrite=False):
    path = os.path.abspath(filename)
    if os.path.exists(path) and not overwrite:
        raise FileExistsError("Output file already exists: %s" % path)
    parent = os.path.dirname(path)
    if parent:
        os.makedirs(parent, exist_ok=True)
    return path


def _python_datetime(value):
    array = np.asarray(value)
    if np.issubdtype(array.dtype, np.datetime64):
        return array.astype("datetime64[us]").tolist()
    if isinstance(value, datetime.datetime):
        return value
    raise TypeError("Unable to convert %r to datetime." % (value,))


def _epoch_parts(dt):
    epoch = datetime.datetime(1970, 1, 1, tzinfo=dt.tzinfo)
    delta = dt - epoch
    total_seconds = int(delta.total_seconds())
    microseconds = int(delta.microseconds)
    return total_seconds, microseconds


def _normalize_ascii_bytes(value, length, fallback):
    text = (str(value).strip() if value is not None else "") or fallback
    return text[:length].ljust(length).encode("ascii", "ignore")


def _available_wsr98d_fields(prd, field_names=None, strict=True):
    available = set(prd.available_fields(range_mode=None))
    if field_names is None:
        selected = [name for name in WSR98D_WRITE_FIELD_SPECS if name in available]
    else:
        selected = []
        for name in field_names:
            if name not in WSR98D_WRITE_FIELD_SPECS:
                if strict:
                    raise ValueError("Field %s is not supported by WSR98D export." % name)
                continue
            if name not in available:
                if strict:
                    raise KeyError(name)
                continue
            selected.append(name)
    if not selected:
        raise ValueError("No exportable WSR98D fields are available.")
    return selected


def _range_geometry(range_values):
    range_values = np.asarray(range_values, dtype=np.float64)
    if range_values.size == 0:
        raise ValueError("Export range vector cannot be empty.")
    if range_values.size > 1:
        spacing = float(range_values[1] - range_values[0])
        if spacing <= 0.0:
            raise ValueError("Range spacing must be positive.")
    else:
        spacing = float(range_values[0]) if float(range_values[0]) > 0.0 else 1.0
    return int(round(float(range_values[0]))), int(round(spacing))


def _select_resolution_range(prd, sweep, candidates):
    for field_name in candidates:
        source_name = field_name
        if hasattr(prd, "resolve_field_name"):
            source_name = prd.resolve_field_name(field_name, sweep=sweep, range_mode=None, required=False)
        if source_name is not None and source_name in prd.available_fields(sweep=sweep, range_mode=None):
            field = prd.get_sweep_field(sweep, source_name, range_mode=None)
            return np.asarray(field["range"].values, dtype=np.float64)
    return np.asarray(prd.fields[sweep]["range"].values, dtype=np.float64)


def _encode_quantized(values, scale, offset, bin_length, missing_code=3):
    max_code = 255 if bin_length == 1 else 65535
    codes = np.round(np.asarray(values, dtype=np.float64) * float(scale) + float(offset))
    codes[~np.isfinite(values)] = missing_code
    codes = np.clip(codes.astype(np.int64), 0, max_code)
    if bin_length == 1:
        return codes.astype(np.uint8).tobytes()
    if bin_length == 2:
        return codes.astype("<u2").tobytes()
    raise ValueError("WSR98D bin length must be 1 or 2.")


class WSR98DBaseData(object):
    """Decode standard WSR-98D dual-polarization base data."""

    def __init__(self, filename, station_lon=None, station_lat=None, station_alt=None):
        """
        :param filename:  radar basedata filename
        :param station_lon:  radar station longitude //units: degree east
        :param station_lat:  radar station latitude //units:degree north
        :param station_alt:  radar station altitude //units: meters
        """
        super(WSR98DBaseData, self).__init__()
        self.filename = filename
        self.station_lon = station_lon
        self.station_lat = station_lat
        self.station_alt = station_alt
        self.fid = _prepare_for_read(self.filename)
        self._check_standard_basedata()
        self.header = self._parse_BaseDataHeader()
        self.radial, self._status, self._azimuth, self._elevation, self._seconds, self._microseconds = \
            self._parse_radial()
        self._scan_time = None
        self._nyquist_velocity = None
        self._unambiguous_range = None
        self.nrays = len(self.radial)
        self.sweep_start_ray_index = np.where((self._status == 0) | (self._status == 3))[0]
        self.sweep_end_ray_index = np.where((self._status == 2) | (self._status == 4))[0]
        self.nsweeps = len(self.sweep_start_ray_index)
        self.fid.close()

    def _check_standard_basedata(self):
        """
        :param fid: file fid
        :return:
        """
        if _read_exact(self.fid, 4, "WSR98D file signature") != b'RSTM':
            raise ValueError("File is not a standard WSR-98D file.")
        self.fid.seek(0, 0)
        return

    def _parse_BaseDataHeader(self):
        BaseDataHeader = {}
        fixed_buf = _read_exact(self.fid, dtype_98D.CutConfigurationBlockPos, "WSR98D fixed header")

        BaseDataHeader['GenericHeader'], _ = _unpack_from_buf(fixed_buf, \
                                                              dtype_98D.GenericHeaderBlockPos,
                                                              dtype_98D.BaseDataHeader['GenericHeaderBlock'])
        BaseDataHeader['SiteConfig'], _ = _unpack_from_buf(fixed_buf, \
                                                           dtype_98D.SiteConfigurationBlockPos,
                                                           dtype_98D.BaseDataHeader['SiteConfigurationBlock'])
        BaseDataHeader['TaskConfig'], _ = _unpack_from_buf(fixed_buf,
                                                           dtype_98D.TaskConfigurationBlockPos,
                                                           dtype_98D.BaseDataHeader['TaskConfigurationBlock'])
        cut_count = _validate_count(
            "WSR98D CutNumber",
            BaseDataHeader['TaskConfig']['CutNumber'],
            minimum=1,
            maximum=MAX_WSR98D_SWEEPS,
        )
        cut_buf = _read_exact(
            self.fid,
            dtype_98D.CutConfigurationBlockSize * cut_count,
            "WSR98D cut configuration block",
        )
        BaseDataHeader['CutConfig'] = np.frombuffer(cut_buf, dtype_98D.BaseDataHeader['CutConfigurationBlock'])
        return BaseDataHeader

    def _parse_radial(self):
        radial = []
        azimuth = []
        elevation = []
        status = []
        seconds = []
        microseconds = []
        buf = _read_all(
            self.fid,
            "WSR98D radial payload",
            max_bytes=WSR98D_MAX_DECODED_PAYLOAD_BYTES,
        )
        buffer_view = memoryview(buf)
        pos = 0
        size = len(buf)
        header_size = _WSR98D_RADIAL_HEADER_STRUCT.size
        radial_keys = _WSR98D_RADIAL_HEADER_KEYS
        radial_struct = _WSR98D_RADIAL_HEADER_STRUCT
        while pos + header_size <= size:
            radial_values = radial_struct.unpack_from(buf, pos)
            pos += header_size
            RadialDict = dict(zip(radial_keys, radial_values))
            moment_num = _validate_count(
                "WSR98D MomentNumber",
                RadialDict['MomentNumber'],
                minimum=0,
                maximum=MAX_WSR98D_MOMENTS_PER_RADIAL,
            )
            RadialDict['fields'], pos = self._parse_radial_single(buffer_view, pos, moment_num)
            radial.append(RadialDict)
            azimuth.append(RadialDict['Azimuth'])
            elevation.append(RadialDict['Elevation'])
            status.append(RadialDict['RadialState'])
            seconds.append(RadialDict['Seconds'])
            microseconds.append(RadialDict['MicroSeconds'])
        if pos != size:
            raise ValueError("WSR98D radial payload contains trailing truncated data.")
        return radial, np.asarray(status), np.asarray(azimuth), np.asarray(elevation), \
            np.asarray(seconds), np.asarray(microseconds)

    def _parse_radial_single(self, buf, pos, moment_num):
        radial_var = {}
        moment_header_size = _WSR98D_MOMENT_HEADER_STRUCT.size
        moment_header_struct = _WSR98D_MOMENT_HEADER_STRUCT
        moment_header_keys = _WSR98D_MOMENT_HEADER_KEYS
        flag_to_product = dtype_98D.flag2Product
        buffer_size = len(buf)
        for _ in range(moment_num):
            if pos + moment_header_size > buffer_size:
                raise ValueError("WSR98D moment header is truncated or malformed.")
            moment_header_values = moment_header_struct.unpack_from(buf, pos)
            Momheader = dict(zip(moment_header_keys, moment_header_values))
            pos += moment_header_size
            data_len = _validate_count(
                "WSR98D moment length",
                Momheader['Length'],
                minimum=0,
                maximum=MAX_WSR98D_MOMENT_DATA_BYTES,
            )
            if pos + data_len > buffer_size:
                raise ValueError("WSR98D moment payload extends beyond the available buffer.")
            Data_buf = buf[pos:pos + data_len]
            pos += data_len
            if Momheader['BinLength'] not in (1, 2):
                raise ValueError("WSR98D moment bin length must be 1 or 2 bytes.")
            if data_len % int(Momheader['BinLength']) != 0:
                raise ValueError("WSR98D moment payload length does not match the bin length.")
            if int(Momheader['Scale']) == 0:
                raise ValueError("WSR98D moment scale cannot be zero.")
            if Momheader['BinLength'] == 1:
                dat_tmp = np.frombuffer(Data_buf, dtype=np.uint8, count=data_len)
            else:
                dat_tmp = np.frombuffer(Data_buf, dtype=np.uint16, count=data_len // 2)
            if Momheader['DataType'] <= 35:
                dat = np.full(dat_tmp.shape, np.nan, dtype=np.float32)
                valid = dat_tmp >= 5
                if np.any(valid):
                    dat[valid] = (dat_tmp[valid].astype(np.float32) - Momheader['Offset']) / Momheader['Scale']
                field_name = flag_to_product.get(int(Momheader['DataType']))
                if field_name is not None:
                    radial_var[field_name] = dat
        return radial_var, pos

    def get_nyquist_velocity(self):
        """Return the per-ray Nyquist velocity."""
        if self._nyquist_velocity is None:
            self._nyquist_velocity = np.repeat(self.header['CutConfig']['NyquistSpeed'], self.get_rays_per_sweep())
        return self._nyquist_velocity

    def get_unambiguous_range(self):
        """Return the per-ray unambiguous range."""
        if self._unambiguous_range is None:
            self._unambiguous_range = np.repeat(self.header['CutConfig']['MaximumRange'], self.get_rays_per_sweep())
        return self._unambiguous_range

    def get_scan_time(self):
        """Return the acquisition time for each ray."""
        if self._scan_time is None:
            self._scan_time = np.array(
                [julian2date_SEC(sec, usec) for sec, usec in zip(self._seconds, self._microseconds)]
            )
        return self._scan_time

    def get_sweep_end_ray_index(self):
        """Return the inclusive end index of each sweep."""
        return self.sweep_end_ray_index

    def get_sweep_start_ray_index(self):
        """Return the start index of each sweep."""
        return self.sweep_start_ray_index

    def get_rays_per_sweep(self):
        """Return the number of rays in each sweep."""
        return self.sweep_end_ray_index - self.sweep_start_ray_index + 1

    def get_azimuth(self):
        """Return the azimuth angle for each ray."""
        return self._azimuth

    def get_elevation(self):
        """Return the elevation angle for each ray."""
        return self._elevation

    def get_latitude_longitude_altitude_frequency(self):
        """Return latitude, longitude, altitude, and radar frequency."""
        lat, lon, alt, frequency = self.header['SiteConfig']['Latitude'], self.header['SiteConfig']['Longitude'], \
               self.header['SiteConfig']['Height'], self.header['SiteConfig']['Frequency'] / 1000.
        if self.station_lon is not None:
            lon = self.station_lon
        if self.station_lat is not None:
            lat = self.station_lat
        if self.station_alt is not None:
            alt = self.station_alt
        return lat, lon, alt, frequency

    def get_scan_type(self):
        if self.header['TaskConfig']['ScanType'] in [0, 1]:
            return "ppi"
        elif self.header['TaskConfig']['ScanType'] in [2, 5]:
            return "rhi"
        elif self.header['TaskConfig']['ScanType'] in [3, 4]:
            return "sector"
        else:
            return "other"
    def get_sitename(self):
        return (self.header['SiteConfig']['SiteName']).decode('UTF-8', 'ignore').strip().strip(b'\x00'.decode())


class WSR98D2NRadar(object):
    """Bridge from raw WSR98D data to an NRadar object."""

    _REFLECTIVITY_LIKE_FIELDS = ("dBZ", "Zc", "dBT")
    _VELOCITY_LIKE_FIELDS = ("V", "Vc", "W", "Wc")
    _RANGE_REFERENCE_FIELDS = ("V", "Vc", "W", "Wc", "dBZ", "Zc", "dBT")

    def __init__(self, WSR98D):
        self.WSR98D = WSR98D
        self.flag_match = np.all(self.WSR98D.header['CutConfig']['LogResolution'] == \
                       self.WSR98D.header['CutConfig']['DopplerResolution'])
        self._dbz_index_cache = {}
        self._sweep_descriptors = self._describe_raw_sweeps()
        self._paired_sweeps = self._pair_split_sweeps()
        self.dBZ_index_alone = np.array([source for source, _ in self._paired_sweeps], dtype=np.int32)
        self.v_index_alone = np.array([target for _, target in self._paired_sweeps], dtype=np.int32)

        for index_with_dbz, index_with_v in self._paired_sweeps:
            self.interp_dBZ(index_with_dbz, index_with_v)

        keep_mask = np.ones(self.WSR98D.nrays, dtype=bool)
        keep_mask[self.get_remove_radial_indices()] = False
        self.radial = [iray for iray, keep in zip(self.WSR98D.radial, keep_mask) if keep]
        self._azimuth = self.WSR98D.get_azimuth()[keep_mask]
        self._elevation = self.WSR98D.get_elevation()[keep_mask]
        self._seconds = self.WSR98D._seconds[keep_mask]
        self._microseconds = self.WSR98D._microseconds[keep_mask]
        self._scan_time = None
        self._nyquist_velocity = None
        self._unambiguous_range = None

        status = self.WSR98D._status[keep_mask]
        self.sweep_start_ray_index = np.where((status == 0) | (status == 3))[0]
        self.sweep_end_ray_index = np.where((status == 2) | (status == 4))[0]
        self.nsweeps = len(self.sweep_start_ray_index)
        self.nrays = len(self.radial)
        self.scan_type = self.WSR98D.get_scan_type()
        self.latitude, self.longitude, self.altitude, self.frequency = \
            self.WSR98D.get_latitude_longitude_altitude_frequency()
        self.header = self.WSR98D.header
        self.header['CutConfig'] = np.delete(self.header['CutConfig'], self.dBZ_index_alone)
        self.bins_per_sweep = self.get_nbins_per_sweep()
        self.max_bins = self.bins_per_sweep.max()
        self.range = self.get_range_per_radial(self.max_bins)
        self.azimuth = self.get_azimuth()
        self.elevation = self.get_elevation()
        self.extended_fields = self._build_extended_fields()
        self.fields = self._get_fields()
        self.sitename = self.WSR98D.get_sitename()

    def _describe_raw_sweeps(self):
        """Classify raw sweeps as reflectivity-only, Doppler-only, or combined."""
        descriptors = []
        for sweep_index, start_idx in enumerate(self.WSR98D.sweep_start_ray_index):
            fields = tuple(self.WSR98D.radial[start_idx]["fields"].keys())
            field_names = set(fields)
            has_reflectivity = any(name in field_names for name in self._REFLECTIVITY_LIKE_FIELDS)
            has_doppler = any(name in field_names for name in self._VELOCITY_LIKE_FIELDS)
            descriptors.append(
                {
                    "index": sweep_index,
                    "fields": fields,
                    "elevation": float(self.WSR98D.header["CutConfig"]["Elevation"][sweep_index]),
                    "reflectivity_only": has_reflectivity and not has_doppler,
                    "doppler_only": has_doppler and not has_reflectivity,
                    "combined": has_reflectivity and has_doppler,
                }
            )
        return descriptors

    def _pair_split_sweeps(self, elevation_tolerance=0.5):
        """Pair reflectivity-only sweeps with the closest compatible Doppler-only sweeps."""
        pending_reflectivity = []
        paired = []
        for descriptor in self._sweep_descriptors:
            sweep_index = descriptor["index"]
            if descriptor["reflectivity_only"]:
                pending_reflectivity.append(sweep_index)
                continue
            if not descriptor["doppler_only"]:
                continue
            match = None
            for candidate in reversed(pending_reflectivity):
                candidate_descriptor = self._sweep_descriptors[candidate]
                if abs(candidate_descriptor["elevation"] - descriptor["elevation"]) <= float(elevation_tolerance):
                    match = candidate
                    break
            if match is None:
                continue
            pending_reflectivity.remove(match)
            paired.append((match, sweep_index))
        return paired

    def get_remove_radial_indices(self):
        """Return the radial indices removed after sweep alignment."""
        dBZ_alone = self.dBZ_index_alone
        index_romove = []
        for isweep in dBZ_alone:
            index_romove.extend(range(self.WSR98D.sweep_start_ray_index[isweep], \
                                      self.WSR98D.sweep_end_ray_index[isweep] + 1))
        return index_romove

    def get_reomve_radial_num(self):
        """Backward-compatible alias for ``get_remove_radial_indices``."""
        return self.get_remove_radial_indices()

    def get_v_idx(self):
        """Return sweep indices that contain Doppler moments without reflectivity."""
        return np.array(
            [descriptor["index"] for descriptor in self._sweep_descriptors if descriptor["doppler_only"]],
            dtype=np.int32,
        )

    def get_dbz_idx(self):
        """Return sweep indices that contain reflectivity without Doppler moments."""
        return np.array(
            [descriptor["index"] for descriptor in self._sweep_descriptors if descriptor["reflectivity_only"]],
            dtype=np.int32,
        )

    def interp_VCP26(self, dBZ_sweep_index, V_sweep_index):
        """
        Handle the VCP26/VCP27 case where reflectivity and velocity sweeps do not align.
        :param dBZ_sweep_index: reflectivity-only sweep indices.
        :param V_sweep_index: Doppler-only sweep indices.
        """
        same_sweeps = min(len(dBZ_sweep_index), len(V_sweep_index))
        for isweep in range(same_sweeps):
            self.interp_dBZ(dBZ_sweep_index[isweep], V_sweep_index[isweep])

    def interp_dBZ(self, field_with_dBZ_num, field_without_dBZ_num):
        """
        Copy reflectivity rays onto the matching Doppler-only sweep by azimuth.
        :param field_with_dBZ_num: source sweep index, 0-based.
        :param field_without_dBZ_num: target sweep index, 0-based.
        """
        azimuth = self.WSR98D.get_azimuth()
        dbz_az = azimuth[self.WSR98D.sweep_start_ray_index[field_with_dBZ_num]: \
                         self.WSR98D.sweep_end_ray_index[field_with_dBZ_num] + 1]
        v_az = azimuth[self.WSR98D.sweep_start_ray_index[field_without_dBZ_num]: \
                       self.WSR98D.sweep_end_ray_index[field_without_dBZ_num] + 1]
        dbz_idx = np.argmin(np.abs(dbz_az.reshape(-1, 1) - v_az.reshape(1, -1)), axis=0) + \
                  self.WSR98D.sweep_start_ray_index[field_with_dBZ_num]

        v_idx = np.arange(self.WSR98D.sweep_start_ray_index[field_without_dBZ_num], \
                          self.WSR98D.sweep_end_ray_index[field_without_dBZ_num] + 1)
        keys = self.WSR98D.radial[self.WSR98D.sweep_start_ray_index[field_with_dBZ_num]]['fields'].keys()
        for ind_dbz, ind_v in zip(dbz_idx, v_idx):
            for ikey in keys:
                self.WSR98D.radial[ind_v]["fields"][ikey] = self.WSR98D.radial[ind_dbz]["fields"][ikey]

    def get_azimuth(self):
        """Return the azimuth angle for each retained ray."""
        return self._azimuth

    def get_elevation(self):
        """Return the elevation angle for each retained ray."""
        return np.where(self._elevation > 180, self._elevation - 360, self._elevation)

    def get_rays_per_sweep(self):
        """Return the number of rays in each retained sweep."""
        return self.sweep_end_ray_index - self.sweep_start_ray_index + 1

    def get_scan_time(self):
        """Return the acquisition time for each retained ray."""
        if self._scan_time is None:
            self._scan_time = np.array(
                [julian2date_SEC(sec, usec) for sec, usec in zip(self._seconds, self._microseconds)]
            )
        return self._scan_time

    def get_nyquist_velocity(self):
        """Return the per-ray Nyquist velocity."""
        if self._nyquist_velocity is None:
            self._nyquist_velocity = np.repeat(self.header['CutConfig']['NyquistSpeed'], self.get_rays_per_sweep())
        return self._nyquist_velocity

    def get_unambiguous_range(self):
        """Return the per-ray unambiguous range."""
        if self._unambiguous_range is None:
            self._unambiguous_range = np.repeat(self.header['CutConfig']['MaximumRange'], self.get_rays_per_sweep())
        return self._unambiguous_range

    def get_sweep_end_ray_index(self):
        """Return the inclusive end index of each retained sweep."""
        return self.sweep_end_ray_index

    def get_sweep_start_ray_index(self):
        """Return the start index of each retained sweep."""
        return self.sweep_start_ray_index

    def get_nbins_per_sweep(self):
        """Return the Doppler gate count for each retained sweep."""
        bins = []
        for idx in self.sweep_start_ray_index:
            radial_fields = self.radial[idx]["fields"]
            selected = None
            for field_name in self._RANGE_REFERENCE_FIELDS:
                selected = radial_fields.get(field_name)
                if selected is not None and selected.size:
                    bins.append(int(selected.size))
                    break
            else:
                sizes = [int(values.size) for values in radial_fields.values() if getattr(values, "size", 0)]
                bins.append(max(sizes) if sizes else 0)
        return np.asarray(bins, dtype=np.int32)

    def get_range_per_radial(self, length, sweep=0):
        """Return Doppler gate-center ranges for a radial of ``length`` bins."""
        Resolution = _decode_wsr98d_resolution(self.header['CutConfig']['DopplerResolution'][int(sweep)])
        return np.linspace(Resolution, Resolution * length, length)

    def get_dbz_range_per_radial(self, length, sweep=0):
        """Return reflectivity gate-center ranges for a radial of ``length`` bins."""
        Resolution = _decode_wsr98d_resolution(self.header['CutConfig']['LogResolution'][int(sweep)])
        return np.linspace(Resolution, Resolution * length, length)

    def _get_fields(self):
        """Assemble the retained fields into dense 2-D arrays."""
        field_keys = []
        seen = set()
        for radial in self.radial:
            for key in radial["fields"].keys():
                if key not in seen:
                    seen.add(key)
                    field_keys.append(key)
        field_keys = tuple(field_keys)
        fields = {ikey: np.full((self.nrays, self.max_bins), np.nan, dtype=np.float64) for ikey in field_keys}
        for sweep_idx, (start, end) in enumerate(zip(self.sweep_start_ray_index, self.sweep_end_ray_index)):
            for iray in range(start, end + 1):
                radial_fields = self.radial[iray]['fields']
                for ikey in field_keys:
                    fields[ikey][iray, :] = self._add_or_del_field(
                        radial_fields.get(ikey),
                        ikey,
                        self.flag_match,
                        sweep_idx=sweep_idx,
                    )
        return fields

    def _build_extended_fields(self):
        """Collect native-range reflectivity for sweeps longer than the aligned grid."""
        extended_sweeps = {}
        scan_time = self.get_scan_time()
        for sweep_idx, (start, end, aligned_bins) in enumerate(
            zip(self.sweep_start_ray_index, self.sweep_end_ray_index, self.bins_per_sweep)
        ):
            native_bins = max(
                (
                    self.radial[iray]["fields"].get("dBZ", np.empty((0,), dtype=np.float32)).size
                    for iray in range(start, end + 1)
                ),
                default=0,
            )
            if native_bins <= int(aligned_bins):
                continue

            data = np.full((end - start + 1, native_bins), np.nan, dtype=np.float32)
            for out_idx, iray in enumerate(range(start, end + 1)):
                dat_ray = self.radial[iray]["fields"].get("dBZ")
                if dat_ray is None or dat_ray.size == 0:
                    continue
                copy_len = min(dat_ray.size, native_bins)
                data[out_idx, :copy_len] = dat_ray[:copy_len].astype(np.float32, copy=False)

            extended_sweeps[sweep_idx] = {
                "data": data,
                "range": self.get_dbz_range_per_radial(native_bins, sweep=sweep_idx).astype(np.float32, copy=False),
                "time": np.asarray(scan_time[start:end + 1]),
                "azimuth": np.asarray(self._azimuth[start:end + 1], dtype=np.float32),
                "elevation": np.asarray(self._elevation[start:end + 1], dtype=np.float32),
                "aligned_bins": int(aligned_bins),
            }

        return {"dBZ": extended_sweeps} if extended_sweeps else {}

    def _get_dbz_resample_index(self, source_length, sweep_idx=0):
        cache_key = (int(sweep_idx), int(source_length))
        cached = self._dbz_index_cache.get(cache_key)
        if cached is not None:
            return cached

        source = self.get_dbz_range_per_radial(source_length, sweep=sweep_idx)
        target = self.range
        right = np.searchsorted(source, target, side="left")
        left = np.clip(right - 1, 0, source_length - 1)
        right = np.clip(right, 0, source_length - 1)
        choose_right = np.abs(source[right] - target) < np.abs(target - source[left])
        nearest = np.where(choose_right, right, left)
        valid = (target >= source[0]) & (target <= source[-1])
        self._dbz_index_cache[cache_key] = (valid, nearest)
        return valid, nearest

    def _add_or_del_field(self, dat_ray, key, flag_match=True, sweep_idx=0):
        """
        Normalize a radial field to the common Doppler-aligned range grid.
        :param dat_ray: radial field data.
        :param key: field name.
        :param flag_match: whether Doppler and reflectivity resolutions match.
        """
        length = self.max_bins
        if dat_ray is None:
            return np.full((length,), np.nan)

        if not flag_match and key == "dBZ":
            valid, nearest = self._get_dbz_resample_index(dat_ray.size, sweep_idx=sweep_idx)
            out = np.full((length,), np.nan)
            out[valid] = dat_ray[nearest[valid]]
            return out

        assert dat_ray.ndim == 1, "check dat_ray"
        if dat_ray.size >= length:
            return dat_ray[:length]
        else:
            out = np.full((length,), np.nan)
            out[:dat_ray.size] = dat_ray
            return out

    def get_nradar_nyquist_speed(self):
        """array shape (nsweeps)"""
        return self.header['CutConfig']['NyquistSpeed']

    def get_NRadar_nyquist_speed(self):
        """Backward-compatible alias for ``get_nradar_nyquist_speed``."""
        return self.get_nradar_nyquist_speed()

    def get_nradar_unambiguous_range(self):
        """array shape (nsweeps)"""
        return self.header['CutConfig']['MaximumRange']

    def get_NRadar_unambiguous_range(self):
        """Backward-compatible alias for ``get_nradar_unambiguous_range``."""
        return self.get_nradar_unambiguous_range()

    def get_fixed_angle(self):
        if self.scan_type == "rhi":
            return self.header['CutConfig']['Azimuth']
        else:
            return self.header['CutConfig']['Elevation']

    def to_prd(self, effective_earth_radius=None):
        """Build an ``NRadar.PRD`` object from the decoded WSR-98D volume."""
        return PRD(fields=self.fields, scan_type=self.scan_type, time=self.get_scan_time(), \
                          range=self.range, azimuth=self.azimuth, elevation=self.elevation, latitude=self.latitude, \
                          longitude=self.longitude, altitude=self.altitude,
                          sweep_start_ray_index=self.sweep_start_ray_index, \
                          sweep_end_ray_index=self.sweep_end_ray_index, fixed_angle=self.get_fixed_angle(), \
                          bins_per_sweep=self.bins_per_sweep, nyquist_velocity=self.get_nradar_nyquist_speed(), \
                          frequency=self.frequency, unambiguous_range=self.get_nradar_unambiguous_range(), \
                          nrays=self.nrays, nsweeps=self.nsweeps, sitename = self.sitename,
                          pyart_radar=None, effective_earth_radius=effective_earth_radius,
                          extended_fields=self.extended_fields,
                          metadata={"original_container": "WSR98D", "radar_name": "WSR98D"})

    def ToPRD(self, effective_earth_radius=None):
        """Backward-compatible alias for ``to_prd``."""
        return self.to_prd(effective_earth_radius=effective_earth_radius)

    def to_pyart_radar(self, effective_earth_radius=None, **kwargs):
        """Export the decoded WSR-98D volume through the PRD Py-ART adapter."""
        return self.to_prd(effective_earth_radius=effective_earth_radius).to_pyart_radar(**kwargs)

    def ToPyartRadar(self, effective_earth_radius=None, **kwargs):
        """Backward-compatible alias for ``to_pyart_radar``."""
        return self.to_pyart_radar(effective_earth_radius=effective_earth_radius, **kwargs)

    def _get_instrument_parameters(self):
        """ Return a dictionary containing instrument parameters. """

        # pulse width
        pulse_width = get_metadata('pulse_width')
        pulse_width['data'] = np.array([self.header["TaskConfig"]['PulseWidth']/10**9,], dtype='float32')  # nanosec->sec
        # assume that the parameters in the first ray represent the beam widths,
        # bandwidth and frequency in the entire volume
        wavelength_hz = self.frequency * 10 ** 9
        # radar_beam_width_h
        radar_beam_width_h = get_metadata('radar_beam_width_h')
        radar_beam_width_h['data'] = np.array([self.header['SiteConfig']['BeamWidthHori'], ], dtype='float32')
        # radar_beam_width_v
        radar_beam_width_v = get_metadata('radar_beam_width_v')
        radar_beam_width_v['data'] = np.array([self.header['SiteConfig']['BeamWidthVert'], ], dtype='float32')
        # frequency
        frequency = get_metadata('frequency')
        frequency['data'] = np.array([wavelength_hz], dtype='float32')
        instrument_parameters = {
            'pulse_width': pulse_width,
            'radar_beam_width_h': radar_beam_width_h,
            'radar_beam_width_v': radar_beam_width_v,
            'frequency': frequency, }

        # nyquist velocity if defined
        nyquist_velocity = get_metadata('nyquist_velocity')
        nyquist_velocity['data'] = self.get_nyquist_velocity()
        instrument_parameters['nyquist_velocity'] = nyquist_velocity
        return instrument_parameters


def write_wsr98d(
    prd,
    filename,
    field_names=None,
    strict=True,
    overwrite=False,
    site_code=None,
    site_name=None,
    task_name=None,
    frequency=None,
    rda_version=722951,
    radar_type=1,
):
    """Write a PRD volume to standard WSR-98D base-data format."""
    _ensure_ppi_prd_volume(prd)
    path = _ensure_output_path(filename, overwrite=overwrite)
    selected_fields = _available_wsr98d_fields(prd, field_names=field_names, strict=strict)

    latitude = float(prd.scan_info["latitude"].values)
    longitude = float(prd.scan_info["longitude"].values)
    altitude = float(prd.scan_info["altitude"].values)
    frequency_ghz = float(frequency) if frequency is not None else float(prd.scan_info["frequency"].values)
    beam_width = np.asarray(prd.scan_info["beam_width"].values, dtype=np.float64)
    fixed_angle = np.asarray(prd.scan_info["fixed_angle"].values, dtype=np.float64)
    rays_per_sweep = np.asarray(prd.scan_info["rays_per_sweep"].values, dtype=np.int32)
    nyquist_velocity = np.asarray(prd.scan_info["nyquist_velocity"].values, dtype=np.float64)
    unambiguous_range = np.asarray(prd.scan_info["unambiguous_range"].values, dtype=np.float64)
    start_time = _python_datetime(prd.scan_info["start_time"].values)

    site_code_bytes = _normalize_ascii_bytes(site_code, 8, "Z0000")
    site_name_bytes = _normalize_ascii_bytes(site_name, 32, getattr(prd, "sitename", "pycwr"))
    task_name_bytes = _normalize_ascii_bytes(task_name, 32, "PYCWR")

    fmt_generic_header = "<" + "".join(item[1] for item in dtype_98D.BaseDataHeader["GenericHeaderBlock"])
    fmt_site_configuration = "<" + "".join(item[1] for item in dtype_98D.BaseDataHeader["SiteConfigurationBlock"])
    fmt_task_configuration = "<" + "".join(item[1] for item in dtype_98D.BaseDataHeader["TaskConfigurationBlock"])
    fmt_radial_header = "<" + "".join(item[1] for item in dtype_98D.RadialHeader())
    fmt_moment_header = "<" + "".join(item[1] for item in dtype_98D.RadialData())

    generic_header = struct.pack(fmt_generic_header, 1297371986, -1, -1, 1, -1, b"")
    site_configuration = struct.pack(
        fmt_site_configuration,
        site_code_bytes,
        site_name_bytes,
        latitude,
        longitude,
        int(round(altitude)),
        int(round(altitude)),
        frequency_ghz * 1000.0,
        float(np.nanmean(beam_width)) if beam_width.size else 1.0,
        float(np.nanmean(beam_width)) if beam_width.size else 1.0,
        int(rda_version),
        int(radar_type),
        b"",
    )
    task_configuration = struct.pack(
        fmt_task_configuration,
        task_name_bytes,
        b"",
        -1,
        0,
        -1,
        int((_python_datetime(start_time) - datetime.datetime(1970, 1, 1, tzinfo=start_time.tzinfo)).total_seconds()),
        int(prd.nsweeps),
        -1.0,
        -1.0,
        -1.0,
        -1.0,
        -1.0,
        -1.0,
        -1.0,
        -1.0,
        -1.0,
        b"",
    )

    cut_config = np.zeros(int(prd.nsweeps), dtype=dtype_98D.BaseDataHeader["CutConfigurationBlock"])
    for sweep in range(int(prd.nsweeps)):
        log_range = _select_resolution_range(prd, sweep, ("dBZ", "Zc", "dBT"))
        dop_range = _select_resolution_range(prd, sweep, ("V", "W", "ZDR", "CC", "PhiDP", "KDP", "SNRH", "SNRV"))
        _, log_resolution = _range_geometry(log_range)
        _, dop_resolution = _range_geometry(dop_range)
        cut_config["ProcessMode"][sweep] = 1
        cut_config["WaveForm"][sweep] = 0
        cut_config["PRF_1"][sweep] = 0.0
        cut_config["PRF_2"][sweep] = 0.0
        cut_config["UnfoldMode"][sweep] = 0
        cut_config["Azimuth"][sweep] = 0.0
        cut_config["Elevation"][sweep] = fixed_angle[sweep]
        cut_config["StartAngle"][sweep] = 0.0
        cut_config["EndAngle"][sweep] = 360.0
        cut_config["AngleResolution"][sweep] = float(beam_width[sweep]) if beam_width.size > sweep else 1.0
        cut_config["ScanSpeed"][sweep] = 0.0
        cut_config["LogResolution"][sweep] = _encode_wsr98d_resolution(log_resolution)
        cut_config["DopplerResolution"][sweep] = _encode_wsr98d_resolution(dop_resolution)
        cut_config["MaximumRange"][sweep] = int(round(unambiguous_range[sweep])) if unambiguous_range.size > sweep else int(round(log_range[-1]))
        cut_config["MaximumRange2"][sweep] = 0
        cut_config["StartRange"][sweep] = int(round(min(log_range[0], dop_range[0])))
        cut_config["Sample_1"][sweep] = int(rays_per_sweep[sweep])
        cut_config["Sample_2"][sweep] = 0
        cut_config["PhaseMode"][sweep] = 0
        cut_config["AtmosphericLoss"][sweep] = 0.0
        cut_config["NyquistSpeed"][sweep] = float(nyquist_velocity[sweep]) if nyquist_velocity.size > sweep else 0.0
        cut_config["MomentsMask"][sweep] = 0
        cut_config["MomentsSizeMask"][sweep] = 0

    with open(path, "wb") as handle:
        handle.write(generic_header)
        handle.write(site_configuration)
        handle.write(task_configuration)
        handle.write(cut_config.tobytes())

        sequence_number = 1
        for sweep in range(int(prd.nsweeps)):
            sweep_dataset = prd.fields[sweep]
            rays = int(rays_per_sweep[sweep])
            for iray in range(rays):
                azimuth = float(sweep_dataset["azimuth"].values[iray])
                elevation = float(sweep_dataset["elevation"].values[iray])
                dt = _python_datetime(sweep_dataset["time"].values[iray])
                seconds, microseconds = _epoch_parts(dt)
                radial_state = 3 if (sweep == 0 and iray == 0) else 4 if (sweep == int(prd.nsweeps) - 1 and iray == rays - 1) else 0 if iray == 0 else 2 if iray == rays - 1 else 1

                moment_headers = []
                moment_payloads = []
                for field_name in selected_fields:
                    if field_name not in prd.available_fields(sweep=sweep, range_mode=None):
                        continue
                    spec = WSR98D_WRITE_FIELD_SPECS[field_name]
                    field = prd.get_sweep_field(sweep, field_name, range_mode=None)
                    values = np.asarray(field.values[iray], dtype=np.float32)
                    data_bytes = _encode_quantized(
                        values,
                        spec["scale"],
                        spec["offset"],
                        spec["bin_length"],
                    )
                    moment_headers.append(
                        struct.pack(
                            fmt_moment_header,
                            spec["data_type"],
                            spec["scale"],
                            spec["offset"],
                            spec["bin_length"],
                            0,
                            len(data_bytes),
                            b"",
                        )
                    )
                    moment_payloads.append(data_bytes)

                length_of_data = sum(len(header) + len(payload) for header, payload in zip(moment_headers, moment_payloads))
                radial_header = struct.pack(
                    fmt_radial_header,
                    radial_state,
                    0,
                    sequence_number,
                    iray + 1,
                    sweep + 1,
                    azimuth,
                    elevation,
                    seconds,
                    microseconds,
                    length_of_data,
                    len(moment_headers),
                    b"",
                )
                handle.write(radial_header)
                for header, payload in zip(moment_headers, moment_payloads):
                    handle.write(header)
                    handle.write(payload)
                sequence_number += 1
    return path



import numpy as np

UINT8 = "B"
UINT16 = "H"
UINT32 = "I"

class SABFormat(object):

    def __init__(self):
        super(SABFormat, self).__init__()
        self.RadialHeaderSize = 128
        self.InfSize = 28

    def RadialHeader(self):
        return (
            ('reserve0', '14s'),
            ('flag', UINT16), ##1-标识雷达数据
            ('reserve1', '12s'),
            ('mSends', UINT32),  # 径向数据收集时间(毫秒,自 00:00 开始)
            ('JulianDate', UINT16),  # 儒略日(Julian)表示,自 1970 年 1 月 1 日开始
            ('URange', UINT16),  # 不模糊距离(表示:数值/10.=千米)
            ('AZ', UINT16),  # 方位角(编码方式:[数值/8.]*[180./4096.]=度)
            ('RadialNumber', UINT16),  # 当前仰角内径向数据序号
            ('RadialStatus', UINT16),  # 径向数据状态
            ('El', UINT16),
            # 仰角 (编码方式:[数值/8.]*[180./4096.]=度)
            ('ElNumber', UINT16),  # 体扫内的仰角数
            ('RangeToFirstGateOfRef', UINT16),  # 反射率数据的第一个距离库的实际距离(单位:米)
            ('RangeToFirstGateOfDop', UINT16),  # 多普勒数据的第一个距离库的实际距离(单位:米)
            ('GateSizeOfReflectivity', UINT16),  # 反射率数据的距离库长(单位:米)
            ('GateSizeOfDoppler', UINT16),  # 多普勒数据的距离库长(单位:米)
            ('GatesNumberOfReflectivity', UINT16),  # 反射率的距离库数
            ('GatesNumberOfDoppler', UINT16),  # 多普勒的距离库数
            ('CutSectorNumber', UINT16),  # 扇区号
            ('CalibrationConst', UINT32),  # 系统订正常数
            ('PtrOfReflectivity', UINT16),  # 反射率数据指针(偏离雷达数据信息头的字节数) 表示第一个反射率数据的位置
            ('PtrOfVelocity', UINT16),  # 速度数据指针(偏离雷达数据信息头的字节数),表示第一个速度数据的位置
            ('PtrOfSpectrumWidth', UINT16),  # 谱宽数据指针(偏离雷达数据信息头的字节数),表示第一个谱宽数据的位置
            ('ResolutionOfVelocity', UINT16),  # 多普勒速度分辨率。 2:表示 0.5 米/秒
            ('VcpNumber', UINT16),  # 体扫(VCP)模式
            ('reserve2', '14s'),  # 保留
            ('Nyquist', UINT16),  # Nyquist 速度(表示:数值/100. = 米/秒)
            ('reserve3', '38s'))

    def RadialData(self, rnumber, vnumber):
        """
        :param rnumber: 反射率的库数
        :param vnumber: 多普勒的库数
        :param radialbins: 整个径向的长度
        :return:
        """
        return np.dtype([('dBZ', 'u1', rnumber),
                ('V', 'u1', vnumber),
                ('W', 'u1', vnumber)])

dtype_sab = SABFormat()